local castbars = ...
table.insert(Event.Addon.Load.End, {function(addon) if (addon == castbars.identifier) then myUI:RegisterModule(castbars) end end, castbars.identifier, "_startup"})
--Command.System.Strict()

local L = castbars.data.language

local ifnil = myUI.Utility.ifnil
local isnil = myUI.Utility.isnil
local rgba = myUI.Utility.rgba
local iif = myUI.Utility.iif
local pprint = myUI.Utility.pprint
local round = myUI.Utility.round
local minmax = myUI.Utility.minmax
local lpad = myUI.Utility.lpad
local trim = myUI.Utility.trim
local isempty = myUI.Utility.isempty
local ifempty = myUI.Utility.ifempty
local nilstr = myUI.Utility.nilstr
local notnil = myUI.Utility.notnil

local colors
local myuicolors = myUI.Colors

local floor = math.floor
local ceil = math.ceil
local abs = math.abs

local sformat = string.format
-- addon --

local _aprofile, _cprofile -- account and character profiles, respectively
local _bars = {}
local _configFrame = nil

local _sensitivity = 10	-- moving/resizing

local _available_units = {
		"player", 
		"player.target", 
		"player.target.target", 
		"player.pet", 
		"player.pet.target",
		"focus", 
		"focus.target",
		"mouseover", 
		"mouseover.target",
	} 


local function _animationUpdate(self, now, elapsed)
	for _, bar in pairs(_bars) do
		if (bar:GetVisible() and not bar.mover:GetVisible()) then
			bar:TimerUpdate(now, elapsed)
		end
	end
end

local function _castUpdate(units) -- units[unitid] = visible	
	local now = Inspect.Time.Frame()
	local uid
	
	for unit, bar in pairs(_bars) do		
		uid = Inspect.Unit.Lookup(bar.unit)
		if (units[uid] ~= nil) then			
			bar:CastUpdate(bar.mover:GetVisible() or units[uid], now)
		end
	end
end


-- Although, almost identical, this differs from _castUpdate in that the value of our unit listing is the unitID, not a visibility flag.
-- It's a bit of a hacky brute force approach, castbar events don't fire when targets change, so there's no other way to make
-- sure your target is/isn't casting a spell if it changed during that spell's cast time.

local function _unitUpdate(units)
	local now = Inspect.Time.Frame()
	
	for unit, bar in pairs(_bars) do		
		bar:CastUpdate(bar.mover:GetVisible() or notnil(Inspect.Unit.Castbar(bar.unit)), now)
	end
end

-- castbar mover --
local function _leftDownM(mover)
	if (mover.allowdrag) then
		local m = Inspect.Mouse()
		if (m) then
			mover.x = m.x
			mover.y = m.y
			mover.dragging = true
		end
	end
end

local function _leftUpM(mover)
	mover.dragging = false
end
local function _mouseMoveM(mover, x, y)	
	if (mover.allowdrag) and (mover.dragging) then
		local deltax = floor(x - mover.x)
		local deltay = floor(y - mover.y)
		
		-- apply sensitivity setting
		deltax = floor(deltax / _sensitivity) * _sensitivity
		deltay = floor(deltay / _sensitivity) * _sensitivity
		
		if (abs(deltax) > 0) or (abs(deltay) > 0) then						
			local xnow = mover:GetLeft() + deltax
			local ynow = mover:GetTop() + deltay
			
			mover:SetPoint("TOPLEFT", UIParent, "TOPLEFT", xnow, ynow)
			
			mover.coordtext:SetText(tostring(math.floor(mover:GetLeft())) .. ", " .. tostring(math.floor(mover:GetTop())))
			
			local def =_bars[mover.unit].def
			def.x = xnow
			def.y = ynow
			
			mover.x = mover.x + deltax
			mover.y = mover.y + deltay
		end
	end
end
local function _rightDownM(mover)
	if (mover:GetVisible()) then
		local s = _sensitivity - 1
		if (s < 1) then s = 10 end		
		_sensitivity = s
		
		for _, uf in pairs(_bars) do
			uf.mover.sensitivity:SetText("+/- " .. tostring(_sensitivity))
		end
	end
end







-- castbar sizer --
local function _leftDownS(sizer)
	if (sizer.allowsize) then
		local m = Inspect.Mouse()
		if (m) then
			sizer.x = m.x
			sizer.y = m.y
			sizer.sizing = true
		end
	end
end
local function _leftUpS(sizer)
	sizer.sizing = false
end

local function _mouseMoveS(sizer, x, y)
	if (sizer.allowsize) and (sizer.sizing) then
		local deltax = floor(x - sizer.x)
		local deltay = floor(y - sizer.y)
		
		-- apply sensitivity setting
		deltax = floor(deltax / _sensitivity) * _sensitivity
		deltay = floor(deltay / _sensitivity) * _sensitivity
		
		if (abs(deltax) > 0) or (abs(deltay) > 0) then
			local mover = sizer:GetParent()
			
			local cxmover = mover:GetWidth()
			local cymover = mover:GetHeight()
			
			local cxnow = cxmover + deltax
			local cynow = cymover + deltay
			
			cxnow = floor(minmax(cxnow, 60, UIParent:GetWidth()))
			cynow = floor(minmax(cynow, 18, UIParent:GetHeight()))
			
			if (cxnow ~= cxmover) or (cynow ~= cymover) then
				mover.sizetext:SetText(tostring(cxnow) .. " x " .. tostring(cynow))
				
				local cb =_bars[mover.unit]
				
				cb.def.cx = cxnow
				cb.def.cy = cynow
				
				cb:Layout(true)
				
				if (cb.group == 1) then					
					-- this should never execute but we'll keep it in case we add raid castbars in the future
					for i=2,20 do
						_bars[string.gsub(cb.unit, "01", lpad(tostring(i), 2, "0"))]:Layout(true)
					end
				end
				
				sizer.x = sizer.x + deltax
				sizer.y = sizer.y	+ deltay
			end
		
		end
	end
end


-- castbar --
local LAYER_CASTBAR = myUI.Layers.UnitFrame + 1
local LAYER_MOVER = LAYER_CASTBAR + 20

local function _getDefinitionCB(unit)
	local def = _aprofile.castbars[unit]
	
	return def
end




local function _onSizeBg(bg)
		bg.cx = bg:GetWidth()	-- pre-compute to save time
end

local function _layoutCB(cb)
	local def = cb.def
	local mover = cb.mover
	
	if (def.enabled) then		
		mover:ClearAll()
		mover:SetSize(ifnil(def.cx, _aprofile.castbar.cx), ifnil(def.cy, _aprofile.castbar.cy))
		mover:SetPoint("TOPLEFT", UIParent, "TOPLEFT", ifnil(def.x, _aprofile.castbar.x), ifnil(def.y, _aprofile.castbar.y))
		mover.sizer:SetVisible(true)
		
		mover.coordtext:SetText(tostring(floor(mover:GetLeft())) .. ", " .. tostring(floor(mover:GetTop())))
		mover.sizetext:SetText(tostring(floor(mover:GetWidth())) .. " x " .. tostring(floor(mover:GetHeight())))
		
		local def = cb.def
		
		cb.bg:ClearAll()
		cb.guage:ClearAll()
		cb.spell:ClearAll()
		cb.remaining:ClearAll()
		
		local margin = ifnil(def.margin, _aprofile.castbar.margin)
		local cxy = cb:GetHeight() - (2 * margin)		
		cxy = iif(cxy < 8, 8, cxy)
		
		cb.bg:SetHeight(cxy)
		cb.bg:SetPoint("LEFTCENTER", cb, "LEFTCENTER", margin, 0)
		cb.bg:SetPoint("RIGHTCENTER", cb, "RIGHTCENTER", -margin, 0)
		cb.bg.cx = cb.bg:GetWidth()
		
		cb.guage:SetPoint("TOPLEFT", cb.bg, "TOPLEFT", 0, 0)
		cb.guage:SetPoint("BOTTOMLEFT", cb.bg, "BOTTOMLEFT", 0, 0)
		
		cb.remaining:SetPoint("RIGHTCENTER", cb.bg, "RIGHTCENTER", -2, 0)
		cb.spell:SetPoint("LEFTCENTER", cb.bg, "LEFTCENTER", 2, 0)
		cb.spell:SetPoint("RIGHTCENTER", cb.remaining, "LEFTCENTER", -4, 0)
		
		cb.remaining:SetFontSize(ifnil(def.fontsizeRemaining, _aprofile.castbar.fontsizeRemaining))		
		cb.spell:SetFontSize(ifnil(def.fontsizeSpell, _aprofile.castbar.fontsizeSpell))
		
		cb:SetBackgroundColor(rgba(colors.border))
		cb.bg:SetBackgroundColor(rgba(colors.bg))
		cb.guage:SetBackgroundColor(rgba(colors.bar))
		cb.spell:SetFontColor(rgba(colors.fgspell))
		cb.remaining:SetFontColor(rgba(colors.fgremaining))
		
		cb:SetVisible(true)	-- temporarily. bars will hide/show themselves automatically when update is implemented
		cb.guage:SetWidth(60)
		cb.spell:SetText("Some spell ")
		cb.remaining:SetText("1.0")
	else
		cb:SetVisible(false)		
	end
end

local function _castUpdateCB(cb, visible, now)
	visible = visible and cb.def.enabled
	
	if (visible) then
		local config = cb.mover:GetVisible()		
		local detail, lastdetail
		if (config) then
			detail = {
				abilityName = "Can o' whoopass",
				begin = now - 0.5,
				duration = 2,
				remaining = 1.5,				
			}
			lastdetail = {}
		else
			detail = Inspect.Unit.Castbar(cb.unit)
			lastdetail = cb.lastdetail
		end
		
		if (detail) then		
			if (detail.abilityName ~= lastdetail.abilityName) then
				cb.spell:SetText(ifnil(detail.abilityName, ""))			
			end
			if (detail.remaining ~= lastdetail.remaining) and (detail.remaining) then
				cb.remaining:SetText(string.format("%.1f", tostring(detail.remaining)))
			end
			if (detail.uninterruptible ~= lastdetail.uninterruptible) then
				if (detail.uninterruptible) then
					cb.guage:SetBackgroundColor(rgba(colors.nointerrupt))
				else
					cb.guage:SetBackgroundColor(rgba(colors.bar))
				end
			end
			
			cb.lastdetail = detail
		else
			cb.lastdetail = {}
		end
		
			
	end
	
	if (cb:GetVisible() ~= visible) then cb:SetVisible(visible) end
end	

local function _timerUpdateCB(cb, now, elapsed)
	if (not cb.mover:GetVisible()) then
		if (cb:GetVisible() and cb.lastdetail) then
			local detail = cb.lastdetail
			local remaining = detail.remaining
			
			if (detail.begin) then
				remaining = detail.begin + detail.duration - now
			else
				remaining = detail.remaining
			end
			
			if (remaining >= 0) then
				local cast = remaining / detail.duration
				
				if (not detail.channeled) then
					cast = 1 - cast
				end
				
				cb.remaining:SetText(sformat("%.1f", tostring(remaining)))
				
				local cxcast = minmax(cast * cb.bg.cx, 0, cb.bg.cx)
				cb.guage:SetWidth(cxcast)
			else
				cb:SetVisible(false)
			end			
		end
	end
end



local function _newCB(unit)
	local cb = myUI:CreateFrame("Frame")
	local mover = myUI:CreateFrame("Frame")
	local sizer = myUI:CreateFrame("Texture", mover)	
	
	cb.mover = mover
	mover.sizer = sizer
	
	mover:SetLayer(LAYER_MOVER)
	mover:SetBackgroundColor(rgba(myuicolors.Mover))
	mover:SetVisible(false)
	
	mover.unittext = myUI:CreateFrame("Text", mover)
	mover.unittext:SetPoint("BOTTOMLEFT", mover, "TOPLEFT", 0, -4)
	mover.unittext:SetFontSize(12); mover.unittext.customFontSize = true	
	mover.unittext:SetBackgroundColor(0.6,0,0,0.6)
	
	mover.coordtext = myUI:CreateFrame("Text", mover)
	mover.coordtext:SetPoint("TOPLEFT", mover.unittext, "TOPRIGHT", 4, 0)
	mover.coordtext:SetFontSize(12); mover.coordtext.customFontSize = true
	mover.coordtext:SetBackgroundColor(0.6,0,0,0.6)
	
	mover.sizetext = myUI:CreateFrame("Text", mover)
	mover.sizetext:SetPoint("TOPLEFT", mover.coordtext, "TOPRIGHT", 4, 0)
	mover.sizetext:SetFontSize(12); mover.sizetext.customFontSize = true
	mover.sizetext:SetBackgroundColor(0.6,0,0,0.6)
	
	mover.sensitivity = myUI:CreateFrame("Text", mover)
	mover.sensitivity:SetPoint("TOPLEFT", mover.sizetext, "TOPRIGHT", 4, 0)
	mover.sensitivity:SetFontSize(12); mover.sensitivity.customFontSize = true
	mover.sensitivity:SetBackgroundColor(0.6,0,0,0.6)
	mover.sensitivity:SetText("+/- " .. tostring(_sensitivity))
	
	mover.x = 0
	mover.y = 0
	mover.dragging = false
	
	mover.Event.LeftDown = _leftDownM
	mover.Event.LeftUp = _leftUpM
	mover.Event.LeftUpoutside = _leftUpM
	mover.Event.MouseMove = _mouseMoveM
	mover.Event.RightDown = _rightDownM
	
	sizer:SetSize(18)
	sizer:SetTexture("myUI", "images/dragger.png")
	sizer:SetPoint("BOTTOMRIGHT", mover, "BOTTOMRIGHT", 0, 0)
	sizer.x = 0
	sizer.y = 0
	sizer.sizing = false
	
	sizer.Event.LeftDown = _leftDownS
	sizer.Event.LeftUp = _leftUpS
	sizer.Event.LeftUpoutside = _leftUpS
	sizer.Event.MouseMove = _mouseMoveS
	
	cb:SetAllPoints(mover)
	cb:SetLayer(LAYER_CASTBAR)
	
	cb.unit = unit	
	mover.unit = unit 
	sizer.unit = unit
	
	mover.allowdrag = true
	sizer.allowsize = true
	
	mover.unittext:SetVisible(true)
	mover.coordtext:SetVisible(true)
	mover.sizetext:SetVisible(true)

	mover.unittext:SetText(unit)
	
	cb.bg = myUI:CreateFrame("Frame", cb)
	cb.bg.Event.Size = _onSizeBg
	
	cb.guage = myUI:CreateFrame("Frame", cb.bg)
	cb.guage:SetLayer(1)
	
	cb.spell = myUI:CreateFrame("Text", cb.bg)
	cb.spell:SetLayer(2)
	cb.spell.customFontSize = true
	
	cb.remaining = myUI:CreateFrame("Text", cb.bg)
	cb.remaining:SetLayer(2)
	cb.remaining.customFontSize = true
	
	local def = _getDefinitionCB(unit)
	
	cb.def = def	
	cb.lastdetail = {}
	
	cb.Layout = _layoutCB
	cb.TimerUpdate = _timerUpdateCB
	cb.CastUpdate = _castUpdateCB
	
	cb:Layout()
	cb:SetVisible(false)
	return cb
end













local function _enableUnit(unit, enabled)
	_aprofile.castbars[unit].enabled = enabled
	
	local bar = _bars[unit]
	if (enabled) and (bar == nil) then
		bar = _newCB(unit)
		_bars[unit] = bar
	end
	if (bar) then 
		bar:Layout()
		bar.mover:SetVisible(enabled)
		if (enabled) then			
			bar:CastUpdate(true, Inspect.Time.Frame())
		end
	end
end
-- public interface methods

function castbars:Initialize()
	
	
end

function castbars:LoadProfile(account, character)
	_aprofile = account
	_cprofile = character

	local adefs = castbars._getAccountDefaults()
	
	myUI.Utility.nilCopy(_aprofile, adefs)
	colors = _aprofile.colors
	
	for unit, def in pairs(_aprofile.castbars) do
		if (def.enabled) then
			if (not _bars[unit]) then
				_bars[unit] = _newCB(unit)
			end
		end
	end
	
	local id = castbars.identifier
	
	
	if (not myUI.Events.AnimationUpdate:IsSubscribed(id .. ".animationUpdate")) then
		myUI.Events.AnimationUpdate:Subscribe(id .. "._animationUpdate", self, _animationUpdate)
	end
	
	myUI:HookEvent(Event.Unit.Availability.Full, _unitUpdate, castbars, id .. "._unitUpdate")
	myUI:HookEvent(Event.Unit.Availability.Partial, _unitUpdate, castbars, id .. "._unitUpdate")
	myUI:HookEvent(Event.Unit.Availability.None, _unitUpdate, castbars, id .. "._unitUpdate")
	
	myUI:HookEvent(Event.Unit.Remove, _unitUpdate, castbars, id .. "._unitUpdate")
	myUI:HookEvent(Event.Unit.Add, _unitUpdate, castbars, id .. "._unitUpdate")
	
	myUI:HookEvent(Event.Unit.Castbar, _castUpdate, castbars, id .. "._castUpdate")
	
	if (_configFrame) then _configFrame:LoadProfile() end	
end

function castbars:GetConfigurationFrame(parent)
	if (not _configFrame) then
		Command.System.Watchdog.Quiet()
		_configFrame = myUI:CreateFrame("Frame", parent)	
		
		local vp = myUI:CreateFrame("myUI_Viewport", _configFrame)
		vp:SetPoint("TOPLEFT", _configFrame, "TOPLEFT", 0, 0)
		vp:SetPoint("BOTTOMRIGHT", _configFrame, "BOTTOMRIGHT", 0, 0)
		
		local content = vp.Content
		content:SetHeight(600)
		--content:SetBackgroundColor(0.6,0,0,0.2)
		
		-- event handlers
		
		local function sliderSizeChanged(slider)
			if (not _configFrame._loading) then
				_aprofile.castbars[slider.unit][slider.setting] = slider:GetPosition()
				
				if (_bars[slider.unit]) then
					_bars[slider.unit]:Layout()
					_bars[slider.unit]:CastUpdate(true, Inspect.Time.Frame())
				end
			end
		end
		
		local function checkEnabled(chk)
			if (not _configFrame._loading) then
				_enableUnit(chk.unit, chk:GetChecked())
			end
		end
		
		-- castbars
		local hdr = myUI:CreateFrame("Text", content)
		hdr:SetPoint("TOPLEFT", content, "TOPLEFT")
		hdr:SetPoint("TOPRIGHT", content, "TOPRIGHT")
		hdr:SetText(" " .. L["Castbars"])
		hdr:SetBackgroundColor(rgba(myuicolors.Hover))
	
		local cols = {180, 80, 140, 140}
	
		henabled = myUI:CreateFrame("Text", content)
		henabled:SetText(L["enabled"])
		henabled:SetPoint("TOPLEFT", hdr, "TOPLEFT", cols[1], 0)
		
		hmargin = myUI:CreateFrame("Text", content)
		hmargin:SetText(L["margin"])
		hmargin:SetPoint("TOPLEFT", henabled, "TOPLEFT", cols[2], 0)
		
		hspellfont = myUI:CreateFrame("Text", content)
		hspellfont:SetText(L["spellfont"])
		hspellfont:SetPoint("TOPLEFT", hmargin, "TOPLEFT", cols[3], 0)
		
		htimerfont = myUI:CreateFrame("Text", content)
		htimerfont:SetText(L["timerfont"])
		htimerfont:SetPoint("TOPLEFT", hspellfont, "TOPLEFT", cols[4], 0)
		
		local rowanchor = hdr
		local unit_controls = {}
		
		for _, unit in ipairs(_available_units) do
			unit_controls[unit] = {}
		
			local utext = myUI:CreateFrame("Text", content)
			utext:SetText(L["u-" .. unit])
			utext:SetPoint("TOPLEFT", rowanchor, "BOTTOMLEFT")
			unit_controls[unit].header = utext
			
			rowanchor = utext
			
			local enabled = myUI:CreateFrame("RiftCheckbox", content)
			enabled:SetPoint("LEFTCENTER", utext, "LEFTCENTER", cols[1] + 8, 0)		
			enabled.setting = "enabled"
			enabled.unit = unit
			enabled.Event.CheckboxChange = checkEnabled
			unit_controls[unit].enabled = enabled
			
			local border = myUI:CreateFrame("RiftSlider", content)
			border:SetWidth(100)
			border:SetRange(0, 8)
			border:SetPoint("LEFTCENTER", enabled, "LEFTCENTER", cols[2]-4, 8)
			border.setting = "margin"
			border.unit = unit
			border.Event.SliderChange = sliderSizeChanged
			unit_controls[unit].margin = border
			
			local fsize = myUI:CreateFrame("RiftSlider", content)
			fsize:SetWidth(100)
			fsize:SetRange(6, 32)
			fsize:SetPoint("LEFTCENTER", border, "LEFTCENTER", cols[3]-4, 0)
			fsize.setting = "fontsizeSpell"
			fsize.unit = unit
			fsize.Event.SliderChange = sliderSizeChanged
			unit_controls[unit].fontsizeSpell = fsize
			
			local ssize = myUI:CreateFrame("RiftSlider", content)
			ssize:SetWidth(100)
			ssize:SetRange(6, 32)
			ssize:SetPoint("LEFTCENTER", fsize, "LEFTCENTER", cols[4]-4, 0)
			ssize.setting = "fontsizeRemaining"
			ssize.unit = unit
			ssize.Event.SliderChange = sliderSizeChanged
			unit_controls[unit].fontsizeRemaining = ssize		
		end
		
		-- colors
		local cpicker = myUI:CreateFrame("myUI_ColorPicker", content)
		cpicker:SetLayer(10)
		cpicker:SetVisible(false)
		cpicker:SetBackgroundColor(0.1,0.1,0.1,1)
		
		_configFrame.ColorPicker = cpicker
		
		local function colorClick(cbox)
			local cpicker = _configFrame.ColorPicker
			
			if (cpicker.box == cbox) then
				cpicker:SetVisible(false)
			else
				cpicker.box = cbox
				cpicker:SetPoint("TOPLEFT", cbox, "BOTTOMLEFT", 0, 2)
				cpicker:SetColor(cbox:GetBackgroundColor())
				cpicker:SetVisible(true)
			end
		end
	
		local function colorPickerColorChanged(cpicker, r, g, b, a)
			local cbox = cpicker.box
			cbox:SetBackgroundColor(r, g, b, a)
			
			_aprofile.colors[cbox.setting] = {round(r,3), round(g,3), round(b,3), round(a,3)}
			
			local now = Inspect.Time.Frame()
			for _, bar in pairs(_bars) do
				bar:Layout()
				bar:CastUpdate(true, now)
			end
		end
	
		
		cpicker.Event.ColorChanged = colorPickerColorChanged
	
		local yoff = rowanchor:GetBottom() - content:GetTop() + 8
	
		hdr = myUI:CreateFrame("Text", content)
		hdr:SetPoint("TOPLEFT", content, "TOPLEFT", 0, yoff)
		hdr:SetPoint("TOPRIGHT", content, "TOPRIGHT", 0, yoff)
		hdr:SetText(L["Colors"])
		hdr:SetBackgroundColor(rgba(myuicolors.Hover))
		
		local cboxes = {}
		unit_controls.colors = cboxes
		
		local function makeCBox(setting)
			local cbox = myUI:CreateFrame("Frame", content)
			cbox:SetSize(24)
			cbox.setting = setting
			cbox.Event.LeftDown = colorClick
			
			local txt = myUI:CreateFrame("Text", content)
			txt:SetPoint("LEFTCENTER", cbox, "RIGHTCENTER", 4, 0)
			txt:SetText(L[setting])
			cboxes[setting] = cbox
			return cbox
		end
		
		cols = {20,220,220}
		
		makeCBox("bar"):SetPoint("TOPLEFT", hdr, "BOTTOMLEFT", cols[1], 8)
		makeCBox("nointerrupt"):SetPoint("TOPLEFT", cboxes.bar, "BOTTOMLEFT", 0, 4)
	
		makeCBox("fgspell"):SetPoint("LEFTCENTER", cboxes.bar, "RIGHTCENTER", cols[2], 0)
		makeCBox("fgremaining"):SetPoint("TOPLEFT", cboxes.fgspell, "BOTTOMLEFT", 0, 4)
		
		makeCBox("bg"):SetPoint("LEFTCENTER", cboxes.fgspell, "RIGHTCENTER", cols[3], 0)
		makeCBox("border"):SetPoint("TOPLEFT", cboxes.bg, "BOTTOMLEFT", 0, 4)
		
		_configFrame.unitControls = unit_controls
		
		_configFrame.LoadProfile = function(self)
			self._loading = true
			
			local ctrls = self.unitControls
			for _, unit in pairs(_available_units) do
				local settings = _aprofile.castbars[unit]
				
				ctrls[unit].enabled:SetChecked(ifnil(settings.enabled, _aprofile.castbar.enabled))
				ctrls[unit].margin:SetPosition(ifnil(settings.margin, _aprofile.castbar.margin))
				ctrls[unit].fontsizeSpell:SetPosition(ifnil(settings.fontsizeSpell, _aprofile.castbar.fontsizeSpell))
				ctrls[unit].fontsizeRemaining:SetPosition(ifnil(settings.fontsizeRemaining, _aprofile.castbar.fontsizeRemaining))				
			end
			
			ctrls = ctrls.colors
			ctrls.bar:SetBackgroundColor(rgba(_aprofile.colors.bar))
			ctrls.nointerrupt:SetBackgroundColor(rgba(_aprofile.colors.nointerrupt))
			ctrls.fgspell:SetBackgroundColor(rgba(_aprofile.colors.fgspell))
			ctrls.fgremaining:SetBackgroundColor(rgba(_aprofile.colors.fgremaining))
			ctrls.bg:SetBackgroundColor(rgba(_aprofile.colors.bg))
			ctrls.border:SetBackgroundColor(rgba(_aprofile.colors.border))
			
			self._loading = false
		end
		
		_configFrame:LoadProfile()
	end
	return _configFrame
end
function castbars:ShowMovers()
	local now = Inspect.Time.Frame()
	
	for _, cb in pairs(_bars) do
		if (cb.def.enabled) then			
			cb.mover:SetVisible(true)
			cb.mover.dragging = false
			cb.mover.sizer.sizing = false
	
			cb:CastUpdate(true, now)
		end
	end
end
function castbars:HideMovers()
	for _, cb in pairs(_bars) do
		cb.mover:SetVisible(false)
		cb.mover.dragging = false
		cb.mover.sizer.sizing = false
		
		cb:CastUpdate(false)
	end
end

--
