local unitframes = ...

table.insert(Event.Addon.Load.End, {function(addon) if (addon == unitframes.identifier) then myUI:RegisterModule(unitframes) end end, unitframes.identifier, "_startup"})
--Command.System.Strict()


local L = unitframes.data.language

local ifnil = myUI.Utility.ifnil
local isnil = myUI.Utility.isnil
local rgba = myUI.Utility.rgba
local iif = myUI.Utility.iif
local pprint = myUI.Utility.pprint
local round = myUI.Utility.round
local minmax = myUI.Utility.minmax
local lpad = myUI.Utility.lpad
local trim = myUI.Utility.trim
local isempty = myUI.Utility.isempty
local ifempty = myUI.Utility.ifempty
local nilstr = myUI.Utility.nilstr
local notnil = myUI.Utility.notnil

local colors
local myuicolors = myUI.Colors

local tinsert = table.insert
local tremove = table.remove

local abs = math.abs
local floor = math.floor
local ceil = math.ceil

local gsub = string.gsub
local sfind = string.find
-- addon --

local _aprofile, _bprofile -- account and character profiles, respectively
local _frames = {}; unitframes.Frames = _frames
local _configFrame = nil
local _ttip
local _sensitivity = 10	-- moving/resizing

local _available_units = {
		"player", 
		"player.target", 
		"player.target.target", 
		"player.pet", 
		"player.pet.target",
		"focus", 
		"focus.target",
		"mouseover", 
		"mouseover.target",
		"group01", 
		"group01.pet"
	} 



local function _secureEnter()
	for _, uf in pairs(_frames) do
		uf:OnCombat(true)
	end
	
	if (_configFrame) then _configFrame:SetCombatControls(true) end
end
local function _secureLeave()
	for _, uf in pairs(_frames) do
		uf:OnCombat(false)
	end
	
	if (_configFrame) then _configFrame:SetCombatControls(false) end
end

local function _unitUpdate()
	for _, uf in pairs(_frames) do
		uf:Update()
	end
end

-- unit frame mover --
local function _leftDownM(mover)
	if (mover.allowdrag) then
		local m = Inspect.Mouse()
		if (m) then
			mover.x = m.x
			mover.y = m.y
			mover.dragging = true
		end
	end
end

local function _leftUpM(mover)
	mover.dragging = false
end
local function _mouseMoveM(mover, x, y)	
	if (mover.allowdrag) and (mover.dragging) then
		local deltax = floor(x - mover.x)
		local deltay = floor(y - mover.y)
		
		-- apply sensitivity setting
		deltax = floor(deltax / _sensitivity) * _sensitivity
		deltay = floor(deltay / _sensitivity) * _sensitivity
		
		
		if (abs(deltax) > 0) or (abs(deltay) > 0) then
			local xnow = mover:GetLeft() + deltax
			local ynow = mover:GetTop() + deltay
			
			mover:SetPoint("TOPLEFT", UIParent, "TOPLEFT", xnow, ynow)
			
			mover.coordtext:SetText(tostring(floor(mover:GetLeft())) .. ", " .. tostring(floor(mover:GetTop())))
			
			local def = _frames[mover.unit].def
			def.x = xnow
			def.y = ynow
			
			mover.x = mover.x + deltax
			mover.y = mover.y + deltay
		end
	end
end
local function _rightDownM(mover)
	if (mover:GetVisible()) then
		local s = _sensitivity - 1
		if (s < 1) then s = 10 end		
		_sensitivity = s
		
		for _, uf in pairs(_frames) do
			uf.mover.sensitivity:SetText("+/- " .. tostring(_sensitivity))
		end
	end
end







-- unit frame sizer --
local function _leftDownS(sizer)
	if (sizer.allowsize) then
		local m = Inspect.Mouse()
		if (m) then
			sizer.x = m.x
			sizer.y = m.y
			sizer.sizing = "size"
		end
	end
end
local function _rightDownS(sizer)
	if (sizer.allowsize) then
		local m = Inspect.Mouse()
		if (m) then 
			sizer.x = m.x
			sizer.y = m.y
			sizer.sizing = "spacing"
		end
	end
end

local function _leftUpS(sizer)
	sizer.sizing = false
end

local function _mouseMoveS(sizer, x, y)
	if (sizer.allowsize) and (sizer.sizing) then
		local deltax = floor(x - sizer.x)
		local deltay = floor(y - sizer.y)
		
		if (sizer.sizing == "spacing") then			
			-- should only happend for 'group01' and 'group01.pet'
			
			deltax = floor(deltax/10)*10	-- make this less sensitive than normal resizing
			deltay = floor(deltay/10)*10
					
			if (abs(deltax) > 0) or (abs(deltay) > 0) then
				local mover = sizer:GetParent()	
				local uf = _frames[mover.unit]
				if (uf.group == 1) then					
					local colspacing = minmax(ifnil(uf.def.colspacing,0) + deltax, 0, 200)
					local rowspacing = minmax(ifnil(uf.def.rowspacing,0) + deltay,0, 200)
					
					uf.def.colspacing = colspacing
					uf.def.rowspacing = rowspacing
					
					for i=1,20 do
						_frames[string.gsub(uf.unit, "01", lpad(tostring(i), 2, "0"))]:Layout(true)
					end
					
					sizer.x = sizer.x + deltax
					sizer.y = sizer.y + deltay
				end
			end
		else
			-- apply sensitivity setting
			deltax = floor(deltax / _sensitivity) * _sensitivity
			deltay = floor(deltay / _sensitivity) * _sensitivity
			
			if (abs(deltax) > 0) or (abs(deltay) > 0) then
				local mover = sizer:GetParent()
				
				local uf = _frames[mover.unit]
				
				local cxmover = mover:GetWidth()
				local cymover = mover:GetHeight()
				
				local cxnow = cxmover + deltax
				local cynow = cymover + deltay
				
				cxnow = floor(minmax(cxnow, 60, UIParent:GetWidth()))
				cynow = floor(minmax(cynow, 40, UIParent:GetHeight()))
				
				if (cxnow ~= cxmover) or (cynow ~= cymover) then
					mover.sizetext:SetText(tostring(cxnow) .. " x " .. tostring(cynow))
					
					
					
					uf.def.cx = cxnow
					uf.def.cy = cynow
					
					uf:Layout(true)
					
					if (uf.group == 1) then						
						for i=2,20 do
							_frames[string.gsub(uf.unit, "01", lpad(tostring(i), 2, "0"))]:Layout(true)
						end
					end
					
					sizer.x = sizer.x + deltax
					sizer.y = sizer.y + deltay				
				end
			
			end
		end
	end
end

-- icons --

local function _mouseInIC(icon)	
	_ttip.text:SetPoint("BOTTOMLEFT", icon, "TOPRIGHT", 8, 0)
	_ttip:SetTip(icon.tooltip)
end

local function _mouseOutIC(icon)
	_ttip:SetTip(nil)
end

-- unit frame --

local LAYER_UF = myUI.Layers.UnitFrame
local LAYER_MOVER = LAYER_UF + 20
local LAYER_TOOLTIP = LAYER_UF + 19
local LAYER_FRAME = LAYER_UF + 8
local LAYER_ICON = 12
local LAYER_BAR =  1
local LAYER_TEXT = 4
local LAYER_TEXTURE = 2
local LAYER_BAR_VALUE = 3
local LAYER_LOS = LAYER_TEXT - 1

local _recycledUnitFrames = {}

local function _showContextMenu(clicker)	
	if (Inspect.Unit.Lookup(clicker.unit)) then
		Command.Unit.Menu(clicker.unit)
	end
end

local function _prepareMacro(macro, unit)
	if (macro == nil) or (macro == "nil") or (macro == "") then
		return nil
	elseif (macro == "%menu") then
		return _showContextMenu
	else
		macro = "suppressmacrofailures;" .. macro
		macro = gsub(gsub(macro, "%%u", "@" .. unit), ";", "\n")
		
		--print(macro)
		
		return macro
	end
end

local function _onCombatUF(frame, combat)
	
	--[[  
	
	Since we're not actually moving/sizing the secure frame, but rather the non-secure frame it's anchored to,
	move/sizing seems to work. However, I'm leaving the code here, commented out, in case Trion decides to
	break that.
	
	local mover = frame.mover
	
	if (combat) then
		mover.allowdrag = false
		mover.sizer.allowsize = false
		mover.dragging = false
		mover.sizer.sizing = false
	else
		local group = frame.group
		mover.allowdrag = (group == nil) or (group == 1)
		mover.sizer.allowsize = mover.allowdrag
	end
	]]--
end



local function _getDefinitionUF(unit)
	unit = gsub(unit, "%d%d", "01")		--  group01 and group01.pet have the settings for all group frames
	local def = _aprofile.frames[unit]
	
	-- make sure some of the sub-tables exist 
	if (def.unitframe == nil) then def.unitframe = {} end
	if (def.portrait == nil) then def.portrait = {} end
	if (def.actions == nil) then def.actions = {} end
	
	if (def.titlebar == nil) then def.titlebar = {} end
	if (def.titlebar_text1 == nil) then def.titlebar_text1 = {} end
	if (def.titlebar_text2 == nil) then def.titlebar_text2 = {} end
	
	if (def.healthbar == nil) then def.healthbar = {} end
	if (def.healthbar_text1 == nil) then def.healthbar_text1 = {} end
	if (def.healthbar_text2 == nil) then def.healthbar_text2 = {} end
	
	if (def.powerbar == nil) then def.powerbar = {} end
	if (def.powerbar_text1 == nil) then def.powerbar_text1 = {} end
	if (def.powerbar_text2 == nil) then def.powerbar_text2 = {} end
	
	if (def.chargebar == nil) then def.chargebar = {} end
	if (def.chargebar_text1 == nil) then def.chargebar_text1 = {} end
	if (def.chargebar_text2 == nil) then def.chargebar_text2 = {} end
	
	if (def.extra_text1 == nil) then def.extra_text1 = {} end
	if (def.extra_text2 == nil) then def.extra_text2 = {} end
	
	if (def.icons == nil) then def.icons = {} end
	if (def.icons.player == nil) then def.icons.player = {} end
	if (def.icons.pvp == nil) then def.icons.pvp = {} end
	if (def.icons.loot == nil) then def.icons.loot = {} end
	if (def.icons.role == nil) then def.icons.role = {} end
	if (def.icons.dead == nil) then def.icons.dead = {} end
	if (def.icons.tier == nil) then def.icons.tier = {} end
	if (def.icons.faction == nil) then def.icons.faction = {} end
	if (def.icons.tagged == nil) then def.icons.tagged = {} end
	if (def.icons.vitality == nil) then def.icons.vitality = {} end
	if (def.icons.ready == nil) then def.icons.ready = {} end
	
	return def
end

local function _initUF(uf, unit, group)
	uf.enabled = false
	
	local mover = uf.mover
	local sizer = mover.sizer
	
	uf.unit = unit	
	mover.unit = unit 
	sizer.unit = unit
	
	uf.isplayer = (unit == "player")
	uf.ispet = (unit == "player.pet")	
	uf.istarget = (unit == "player.target")
	uf.ismouseover = (string.find(unit, "mouseover") == 1)
	uf.group = group
	
	uf.updateCount = 0
	
	local canmove = (group == nil) or (group == 1)
	
	mover.allowdrag = canmove
	sizer.allowsize = canmove

	mover.unittext:SetVisible(canmove)
	mover.coordtext:SetVisible(canmove)
	mover.sizetext:SetVisible(canmove)
	mover.sensitivity:SetVisible(canmove)
		
	mover.unittext:SetText(unit)
	
	local def = _getDefinitionUF(unit)
	
	uf.def = def	
	uf.lastDetail = {}
	uf.enabled = true
	
	uf.clicker.unit = unit
	
	local events = uf.clicker.Event
	if (uf.ismouseover) then
		events.LeftDown = nil
		events.RightDown = nil
		events.MiddleDown = nil
		events.Mouse4Down = nil
		events.Mouse5Down = nil		
	else
		events.LeftDown = _prepareMacro(ifnil(def.actions["LeftDown"], _aprofile.actions["LeftDown"]), unit)
		events.RightDown = _prepareMacro(ifnil(def.actions["RightDown"], _aprofile.actions["RightDown"]), unit)
		events.MiddleDown = _prepareMacro(ifnil(def.actions["MIddleDown"], _aprofile.actions["MiddleDown"]), unit)
		events.Mouse4Down = _prepareMacro(ifnil(def.actions["Mouse4Down"], _aprofile.actions["Mouse4Down"]), unit)
		events.Mouse5Down = _prepareMacro(ifnil(def.actions["Mouse5Down"], _aprofile.actions["Mouse5Down"]), unit)
	end
	
	if (uf.ismouseover) then
		uf.clicker:SetMouseoverUnit(nil)
	else
		uf.clicker:SetMouseoverUnit(unit)
	end
	
	if (uf.group ~= nil) then
		sizer.Event.RightUp = _leftUpS
		sizer.Event.RightUpoutside = _leftUpS
		sizer.Event.RightDown = _rightDownS
	else
		sizer.Event.RightUp = nil
		sizer.Event.RightUpoutside = nil
		sizer.Event.RightDown = nil
	end
	
	uf.chargeShown = nil
	
	uf:Layout(true)	
	
	uf.enabled = true
end
local function _updateMacrosUF(uf)
	if not Inspect.System.Secure() then
		local events = uf.clicker.Event
		local unit = uf.unit
		local def = _getDefinitionUF(unit)
		
		if (uf.ismouseover) then
			events.LeftDown = nil
			events.RightDown = nil
			events.MiddleDown = nil
			events.Mouse4Down = nil
			events.Mouse5Down = nil		
		else
			events.LeftDown = _prepareMacro(ifnil(def.actions["LeftDown"], _aprofile.actions["LeftDown"]), unit)
			events.RightDown = _prepareMacro(ifnil(def.actions["RightDown"], _aprofile.actions["RightDown"]), unit)
			events.MiddleDown = _prepareMacro(ifnil(def.actions["MIddleDown"], _aprofile.actions["MiddleDown"]), unit)
			events.Mouse4Down = _prepareMacro(ifnil(def.actions["Mouse4Down"], _aprofile.actions["Mouse4Down"]), unit)
			events.Mouse5Down = _prepareMacro(ifnil(def.actions["Mouse5Down"], _aprofile.actions["Mouse5Down"]), unit)
		end
		
		return true 
	end
end
local function _disposeUF(uf)
	-- Here's a dirty little trick. Since we can't set our "clicker" frame to not visible in combat due to it being restricted, we move
	-- its container (mover) offscreen, effectively hiding it
	uf.mover:ClearAll()
	uf.mover:SetPoint("TOPLEFT", UIParent, "TOPRIGHT", 100, 0)

	uf:SetVisible(false)
	uf.mover:SetVisible(false)
	
	uf.mover.dragging= false
	uf.mover.sizer.sizing = false
	
	uf.enabled = false
end


local function _addIndicator(t, indicator, ctrl)
	if (indicator) and (indicator ~= "nil") then
		if (t[indicator] == nil) then t[indicator] = {} end
		tinsert(t[indicator], ctrl)
	end
end
local function _addTextValue(t, value, ctrl)
	if (value) and (value ~= "nil") then
		if (t[value] == nil) then t[value] = {} end
		tinsert(t[value], ctrl)
	end
end
local function _setupBgColor(def, ctrl, ctrlSetting, indicators, indicators2, def2)
	if (def2 == nil) then def2 = _aprofile end

	ctrl.SetColor = ctrl.SetBackgroundColor
	ctrl.color = ifnil(def[ctrlSetting].bgcolor, def2[ctrlSetting].bgcolor)
	ctrl:SetColor(rgba(ctrl.color))
	
	_addIndicator(indicators2, ifnil(def[ctrlSetting].indicator2, def2[ctrlSetting].indicator2), ctrl)
	_addIndicator(indicators, ifnil(def[ctrlSetting].indicator, def2[ctrlSetting].indicator), ctrl)
end

local function _setupFgColor(def, ctrl, ctrlSetting, indicators, indicators2, text_values)
	ctrl.SetColor = ctrl.SetFontColor
	ctrl.color = ifnil(def[ctrlSetting].fgcolor, _aprofile[ctrlSetting].fgcolor)
	ctrl:SetColor(rgba(ctrl.color))
	
	ctrl:SetFontSize(ifnil(def[ctrlSetting].fontsize, _aprofile[ctrlSetting].fontsize))
		
	_addIndicator(indicators2, ifnil(def[ctrlSetting].indicator2, _aprofile[ctrlSetting].indicator2), ctrl)
	_addIndicator(indicators, ifnil(def[ctrlSetting].indicator, _aprofile[ctrlSetting].indicator), ctrl)
	
	
	_addTextValue(text_values, ifnil(def[ctrlSetting].text, _aprofile[ctrlSetting].text), ctrl)
	
end


local function _layoutIcon(uf, iconName)	
	local d1 = uf.def.icons[iconName]
	local d2 = _aprofile.icons[iconName]
	
	local show = ifnil(d1.show, d2.show)
	local icon = uf.icons[iconName]
	
	if (icon) then
		icon.hidden = not show		
	elseif (show) then		
		icon = myUI:CreateFrame("Texture", uf)
		icon:SetMouseMasking("limited")
		icon:SetLayer(LAYER_ICON)		
		uf.icons[iconName] = icon		
	end
	
	if (icon) then
		if (icon:GetVisible()) then icon:SetVisible(false) end
		if (not icon.hidden) then
			local anchor = ifnil(d1.anchor, d2.anchor)
			icon:ClearAll()
			icon:SetSize(ifnil(d1.cxy, d2.cxy))
			icon:SetPoint("CENTER", uf, anchor, ifnil(d1.xoff, d2.xoff), ifnil(d1.yoff, d2.yoff))	
			
			local tip = ifnil(d1.tooltip, d2.tooltip)
			if (tip == "nil") then tip = nil end
			icon.tooltip = tip
			
			if (tip) then
				icon.Event.MouseIn = _mouseInIC
				icon.Event.MouseOut = _mouseOutIC				
			end
		else
			icon.Event.MouseIn = nil
			icon.Event.MouseOut = nil
		end
	end	
end
local function _setIcon(uf, iconName, textureName)
	local icon = uf.icons[iconName]
	
	if (icon) then
		if (textureName) and (not icon.hidden) then
			if (textureName ~= icon.textureName) then
				icon.textureName = textureName
			
				local tx = _aprofile.textures[textureName]
				if (tx) then
					icon:SetTexture(tx.src, tx.file)
				end
			end
			
			if (not icon:GetVisible()) then
				icon:SetVisible(true)
			end
		else			
			if (icon:GetVisible()) then
				icon:SetVisible(false)
			end
		end
	end
end

local function _layoutUF(uf, doExternal, barsOnly)
		
	local def = uf.def
	local mover = uf.mover
	
	-- note, mover is not 'restricted' and is the actual location/size controller, with the unit frame anchored to it
	if (doExternal) then
		mover:ClearAll()		
		mover:SetSize(def.cx, def.cy)
		
		if (uf.group == nil) or (uf.group == 1) then		
			mover:SetPoint("TOPLEFT", UIParent, "TOPLEFT", def.x, def.y)
			mover.sizer:SetVisible(true)
			
			
			mover.coordtext:SetText(tostring(floor(mover:GetLeft())) .. ", " .. tostring(floor(mover:GetTop())))
			mover.sizetext:SetText(tostring(floor(mover:GetWidth())) .. " x " .. tostring(floor(mover:GetHeight())))
		else
			local group = uf.group
			local col = math.modf((group-1)/5)
			local row = group - ((col) * 5) - 1
			
			local xoff = (def.cx * col) + (def.colspacing * col) 
			local yoff = (def.cy * row) + (def.rowspacing * row) 
			
			local anchor = _frames[string.gsub(uf.unit, "%d%d", "01")]
			mover:SetPoint("TOPLEFT", anchor, "TOPLEFT", xoff, yoff)
			mover.sizer:SetVisible(false)
		end
		
	end
	
	
	local indicators = {}
	local indicators2 = {}
	
	local text_values = {}
	
	local margin = ifnil(def.unitframe.margin, _aprofile.unitframe.margin)
	
	if (not barsOnly) then		
		_setupBgColor(def, uf, "unitframe", indicators, indicators2)
	end
	
	-- portrait
	local showPortrait = ifnil(def.portrait.show, _aprofile.portrait.show)
	if (not barsOnly) then		
		uf.portrait:ClearAll()
		uf.portrait:SetVisible(showPortrait)
		
		if (showPortrait) then
			local cxy = uf:GetHeight() - (2 * margin)
			local m = ifnil(def.portrait.margin, _aprofile.portrait.margin)
			
			uf.portrait:ClearAll()
			uf.portrait:SetSize(cxy)
			uf.portrait.texture:SetSize(cxy - (2 * m))
			
			local anchor = ifnil(def.portrait.align, _aprofile.portrait.align)
			anchor = iif(anchor == "LEFT", "LEFTCENTER", "RIGHTCENTER")
			local offx = margin * (iif(anchor == "LEFTCENTER", 1, -1))
			
			uf.portrait:SetPoint(anchor, uf, anchor, offx, 0)
			uf.portrait.texture:SetTexture(_aprofile.textures.unknown.src, _aprofile.textures.unknown.file)
			
			_setupBgColor(def, uf.portrait, "portrait", indicators, indicators2)		
		end
	end
	
	local cyratio = 0
	local cytitlebar = 0; local cyhealthbar = 0; local cypowerbar = 0; local cychargebar = 0
	local m
	
	-- titlebar
	local showTitlebar = ifnil(def.titlebar.show, _aprofile.titlebar.show)
	
	if (showTitlebar) then
		m = ifnil(def.titlebar.margin, _aprofile.titlebar.margin)
		
		cytitlebar = ifnil(def.titlebar.cyratio, _aprofile.titlebar.cyratio)
		cyratio = cyratio + cytitlebar
				
			
		if (not barsOnly) then
			uf.titlebar:ClearAll()
			uf.titlebar:SetVisible(showTitlebar)
		
			m = ifnil(def.portrait.margin, _aprofile.portrait.margin)
		
			if (showPortrait) then
				
				if (ifnil(def.portrait.align, _aprofile.portrait.align) == "LEFT") then
					uf.titlebar:SetPoint("TOPLEFT", uf.portrait, "TOPRIGHT", m, 0)
					uf.titlebar:SetPoint("TOPRIGHT", uf, "TOPRIGHT", -margin, margin)
				else
					uf.titlebar:SetPoint("TOPLEFT", uf, "TOPLEFT", margin, margin)
					uf.titlebar:SetPoint("TOPRIGHT", uf.portrait, "TOPLEFT", -m, 0)
				end
			else
				uf.titlebar:SetPoint("TOPLEFT", uf, "TOPLEFT", margin, margin)
				uf.titlebar:SetPoint("TOPRIGHT", uf, "TOPRIGHT", -margin, margin)			
			end

			_setupBgColor(def, uf.titlebar, "titlebar", indicators, indicators2)
			_setupFgColor(def, uf.titlebar.text1, "titlebar_text1", indicators, indicators2, text_values)
			_setupFgColor(def, uf.titlebar.text2, "titlebar_text2", indicators, indicators2, text_values)
			
			uf.titlebar.value:ClearAll()
			uf.titlebar.value:SetPoint("TOPLEFT", uf.titlebar, "TOPLEFT", m, m)
			uf.titlebar.value:SetPoint("BOTTOMLEFT", uf.titlebar, "BOTTOMLEFT", m, -m)
			
			uf.titlebar.text1:ClearAll(); uf.titlebar.text2:ClearAll()
			
			local x1 = ifnil(def.titlebar_text1.xoff, _aprofile.titlebar_text1.xoff)
			local x2 = ifnil(def.titlebar_text2.xoff, _aprofile.titlebar_text2.xoff)
			local y1 = ifnil(def.titlebar_text1.yoff, _aprofile.titlebar_text1.yoff)
			local y2 = ifnil(def.titlebar_text2.yoff, _aprofile.titlebar_text2.yoff)
			
			uf.titlebar.text2:SetPoint("RIGHTCENTER", uf.titlebar, "RIGHTCENTER", -m - 2 + x2, y2)
			uf.titlebar.text1:SetPoint("LEFTCENTER", uf.titlebar, "LEFTCENTER", m + 2 + x1, y1)
			uf.titlebar.text1:SetPoint("RIGHTCENTER", uf.titlebar.text2, "LEFTCENTER", -8, y1 - y2)
			
		end
	else
		if (uf.titlebar) and (uf.titlebar:GetVisible()) then uf.titlebar:SetVisible(false) end		
	end
	
	
	-- healthbar	
	
	cyhealthbar = ifnil(def.healthbar.cyratio, _aprofile.healthbar.cyratio)
	cyratio = cyratio + cyhealthbar
	
	if (not barsOnly) then	
		uf.healthbar:ClearAll()
		m = ifnil(def.healthbar.margin, _aprofile.healthbar.margin)
		
		if (showTitlebar) then
			uf.healthbar:SetPoint("TOPLEFT", uf.titlebar, "BOTTOMLEFT", 0, 0)
			uf.healthbar:SetPoint("TOPRIGHT", uf.titlebar, "BOTTOMRIGHT", 0, 0)
		elseif (showPortrait) then
			m = ifnil(def.portrait.margin, _aprofile.portrait.margin)
			if (ifnil(def.portrait.align, _aprofile.portrait.align) == "LEFT") then
				uf.healthbar:SetPoint("TOPLEFT", uf.portrait, "TOPRIGHT", m, 0)
				uf.healthbar:SetPoint("TOPRIGHT", uf, "TOPRIGHT", -margin, margin)
			else
				uf.healthbar:SetPoint("TOPLEFT", uf, "TOPLEFT", margin, margin)
				uf.healthbar:SetPoint("TOPRIGHT", uf.portrait, "TOPLEFT", -m, 0)
			end
		else
			uf.healthbar:SetPoint("TOPLEFT", uf, "TOPLEFT", margin, margin)
			uf.healthbar:SetPoint("TOPRIGHT", uf, "TOPRIGHT", -margin, margin)
		end
		
		_setupFgColor(def, uf.healthbar.text1, "healthbar_text1", indicators, indicators2, text_values)
		_setupFgColor(def, uf.healthbar.text2, "healthbar_text2", indicators, indicators2, text_values)
		_setupBgColor(def, uf.healthbar, "healthbar", indicators, indicators2)
		
		uf.healthbar.value:SetPoint("TOPLEFT", uf.healthbar, "TOPLEFT", m, m)
		uf.healthbar.value:SetPoint("BOTTOMLEFT", uf.healthbar, "BOTTOMLEFT", m, -m)
		uf.healthbar.cx = uf.healthbar:GetWidth() - (2*m)	-- precompute this value
		
		uf.healthbar.value:SetBackgroundColor(rgba(_aprofile.colors.bars.health))
		
		local x1 = ifnil(def.healthbar_text1.xoff, _aprofile.healthbar_text1.xoff)
		local x2 = ifnil(def.healthbar_text2.xoff, _aprofile.healthbar_text2.xoff)
		local y1 = ifnil(def.healthbar_text1.yoff, _aprofile.healthbar_text1.yoff)
		local y2 = ifnil(def.healthbar_text2.yoff, _aprofile.healthbar_text2.yoff)
		
		uf.healthbar.text1:ClearAll(); uf.healthbar.text2:ClearAll()
		uf.healthbar.text2:SetPoint("RIGHTCENTER", uf.healthbar, "RIGHTCENTER", -m - 2 + x2, y2)
		uf.healthbar.text1:SetPoint("LEFTCENTER", uf.healthbar, "LEFTCENTER", m + 2 + x1, y1)
		uf.healthbar.text1:SetPoint("RIGHTCENTER", uf.healthbar.text2, "LEFTCENTER", -8, y1 - y2)
	end
	
	-- extra text
	if (not barsOnly) then
		_setupFgColor(def, uf.extra.text1, "extra_text1", indicators, indicators2, text_values)
		_setupFgColor(def, uf.extra.text2, "extra_text2", indicators, indicators2, text_values)
		
		local x1 = ifnil(def.extra_text1.xoff, _aprofile.extra_text1.xoff)
		local x2 = ifnil(def.extra_text2.xoff, _aprofile.extra_text2.xoff)
		local y1 = ifnil(def.extra_text1.yoff, _aprofile.extra_text1.yoff)
		local y2 = ifnil(def.extra_text2.yoff, _aprofile.extra_text2.xoff)
		
		uf.extra.text1:ClearAll(); uf.extra.text2:ClearAll()
		uf.extra.text1:SetPoint("LEFTCENTER", uf, "LEFTCENTER", margin + x1, y1)
		uf.extra.text2:SetPoint("RIGHTCENTER", uf, "RIGHTCENTER", -margin + x2, y2)
	end
	
	-- powerbar
	
	cypowerbar = ifnil(def.powerbar.cyratio, _aprofile.powerbar.cyratio)
	cyratio = cyratio + cypowerbar	
	
	if (not barsOnly) then
		m = ifnil(def.powerbar.margin, _aprofile.powerbar.margin)
	
		uf.powerbar:ClearAll()
		uf.powerbar:SetPoint("TOPLEFT", uf.healthbar, "BOTTOMLEFT")
		uf.powerbar:SetPoint("TOPRIGHT", uf.healthbar, "BOTTOMRIGHT")
		
		_setupFgColor(def, uf.powerbar.text1, "powerbar_text1", indicators, indicators2, text_values)
		_setupFgColor(def, uf.powerbar.text2, "powerbar_text2", indicators, indicators2, text_values)
		_setupBgColor(def, uf.powerbar, "powerbar", indicators, indicators2)
		
		uf.powerbar.value:SetPoint("TOPLEFT", uf.powerbar, "TOPLEFT", m, m)
		uf.powerbar.value:SetPoint("BOTTOMLEFT", uf.powerbar, "BOTTOMLEFT", m, -m)
		uf.powerbar.cx = uf.powerbar:GetWidth() - (2*m)	-- precompute
		
		--uf.powerbar.value:SetBackgroundColor(0,0,0,0) -- leave transparent until the first update
		
		local x1 = ifnil(def.powerbar_text1.xoff, _aprofile.powerbar_text1.xoff)
		local x2 = ifnil(def.powerbar_text2.xoff, _aprofile.powerbar_text2.xoff)
		local y1 = ifnil(def.powerbar_text1.yoff, _aprofile.powerbar_text1.yoff)
		local y2 = ifnil(def.powerbar_text2.yoff, _aprofile.powerbar_text2.yoff)
		
		uf.powerbar.text1:ClearAll(); uf.powerbar.text2:ClearAll()
		uf.powerbar.text2:SetPoint("RIGHTCENTER", uf.powerbar, "RIGHTCENTER", -m - 2 + x2, y2)
		uf.powerbar.text1:SetPoint("LEFTCENTER", uf.powerbar, "LEFTCENTER", m+2 + x1, y1)
		uf.powerbar.text1:SetPoint("RIGHTCENTER", uf.powerbar.text2, "LEFTCENTER", -8, y1 - y2)
	end
	-- chargebar
	
	local showCharge = (uf.unit == "player") and ifnil(def.chargebar.show, _aprofile.chargebar.show)
	if (showCharge) then		
		local detail = Inspect.Unit.Detail("player")
		if (detail) and (detail.calling == "mage") then			
			cychargebar = ifnil(def.chargebar.cyratio, _aprofile.chargebar.cyratio)
			cyratio = cyratio + cychargebar

			if (not uf.chargebar) then
				uf.chargebar = myUI:CreateFrame("Frame", uf)
				uf.chargebar.text1 = myUI:CreateFrame("Text", uf.chargebar); uf.chargebar.text1.customFontSize = true; uf.chargebar.text1:SetLayer(5)
				uf.chargebar.text2 = myUI:CreateFrame("Text", uf.chargebar); uf.chargebar.text2.customFontSize = true; uf.chargebar.text2:SetLayer(4)
				uf.chargebar.value = myUI:CreateFrame("Texture", uf.chargebar); uf.chargebar.value:SetLayer(1)
			end
			
			if  (not barsOnly) or (not uf.chargeShown) then			
				uf.chargebar:ClearAll()
				
				uf.chargebar:SetPoint("TOPLEFT", uf.powerbar, "BOTTOMLEFT")
				uf.chargebar:SetPoint("TOPRIGHT", uf.powerbar, "BOTTOMRIGHT")
				
				_setupFgColor(def, uf.chargebar.text1, "chargebar_text1", indicators, indicators2, text_values)
				_setupFgColor(def, uf.chargebar.text2, "chargebar_text2", indicators, indicators2, text_values)
				_setupBgColor(def, uf.chargebar, "chargebar", indicators, indicators2)
				
				m = ifnil(def.chargebar.margin, _aprofile.chargebar.margin)
				
				uf.chargebar.value:SetPoint("TOPLEFT", uf.chargebar, "TOPLEFT", m, m)
				uf.chargebar.value:SetPoint("BOTTOMLEFT", uf.chargebar, "BOTTOMLEFT", m, -m)
				uf.chargebar.cx = uf.chargebar:GetWidth() - (2*m) -- preccomputer
				
				uf.chargebar.value:SetBackgroundColor(rgba(colors.bars.charge))
				
				local x1 = ifnil(def.chargebar_text1.xoff, _aprofile.chargebar_text1.xoff)
				local x2 = ifnil(def.chargebar_text2.xoff, _aprofile.chargebar_text2.xoff)
				local y1 = ifnil(def.chargebar_text1.yoff, _aprofile.chargebar_text1.yoff)
				local y2 = ifnil(def.chargebar_text2.yoff, _aprofile.chargebar_text2.yoff)
				
				uf.chargebar.text1:ClearAll(); uf.chargebar.text2:ClearAll()
				uf.chargebar.text2:SetPoint("RIGHTCENTER", uf.chargebar, "RIGHTCENTER", -m - 2 + x2, y2)
				uf.chargebar.text1:SetPoint("LEFTCENTER", uf.chargebar, "LEFTCENTER", m+2 + x1, y1)
				uf.chargebar.text1:SetPoint("RIGHTCENTER", uf.chargebar.text2, "LEFTCENTER", -8, y1 - y2)
				
				uf.chargeShown = true
			end
		else
			showCharge = false
		end
	end
	
	
	-- set relative bar heights
	local cyuf = uf:GetHeight() - (2 * margin)
	
	if (showTitlebar) then
		uf.titlebar:SetHeight(cyuf * (cytitlebar/cyratio))			
	end
	
	uf.healthbar:SetHeight(cyuf * (cyhealthbar/cyratio))
	uf.powerbar:SetHeight(cyuf * (cypowerbar/cyratio))
	
	if (showCharge) then		
		uf.chargebar:SetHeight(cyuf * (cychargebar/cyratio))
	end
	
	
	
	if (not barsOnly) then
		-- icons
		_layoutIcon(uf, "player")
		_layoutIcon(uf, "pvp")
		_layoutIcon(uf, "loot")
		_layoutIcon(uf, "role")
		_layoutIcon(uf, "dead")
		_layoutIcon(uf, "tier")
		_layoutIcon(uf, "faction")
		_layoutIcon(uf, "tagged")
		_layoutIcon(uf, "vitality")
		_layoutIcon(uf, "ready")
		-- layers
		if (uf.titlebar) then 
			uf.titlebar.text1:SetLayer(5); 
			uf.titlebar.text2:SetLayer(4); 
		end
		uf.healthbar.text1:SetLayer(4); uf.healthbar.text2:SetLayer(5); uf.healthbar.value:SetLayer(2)
		uf.healthbar.texture:SetLayer(uf.healthbar.value:GetLayer()-1)
		uf.powerbar.text1:SetLayer(4); uf.powerbar.text2:SetLayer(5); uf.powerbar.value:SetLayer(2)
		uf.powerbar.texture:SetLayer(uf.powerbar.value:GetLayer()-1)
		
		if (uf.chargebar) then uf.chargebar.text1:SetLayer(4); uf.chargebar.text2:SetLayer(5); uf.chargebar.value:SetLayer(2) end
		
		
	end
	
	uf.indicators = indicators
	uf.indicators2 = indicators2
	
	uf.text_values = text_values
	
	if (doExternal) then uf.reset = true end
end
local function _resetUF(uf)
	uf.reset = true
end

local function _applyIndicators(t1, t2, test, clrTrue)
	--[[
	if (t1) then
		if (test) then			
			for _, ctrl in ipairs(t1) do ctrl:SetColor(rgba(clrTrue)) end
		else
			for _, ctrl in ipairs(t1) do ctrl:SetColor(rgba(ctrl.color)) end			
		end
	end
	]]--
	local ctrls = {}
	
	if (t2) then
		if (test) then
			for _, ctrl in ipairs(t2) do
				ctrls[ctrl] = clrTrue
			end
		else
			for _, ctrl in ipairs(t2) do
				ctrls[ctrl] = ctrl.color
			end
		end
	end
	if (t1) then
		if (test) then
			for _, ctrl in ipairs(t1) do
				ctrls[ctrl] = clrTrue
			end
		else
			for _, ctrl in ipairs(t1) do
				if (ctrls[ctrl] == nil) then -- very important check. w/o it we'd overwrite anything set by out secondy indicator
					ctrls[ctrl] = ctrl.color
				end
			end
		end
	end
	
	-- update all controls that have this indicator
	for ctrl, clr in pairs(ctrls) do ctrl:SetColor(rgba(clr)) end
	
	
end
local function _applyValues(t, tag, value)
	if (t) then
		value = ifnil(value, "")
		for _, ctrl in ipairs(t) do 
			if (value ~= ((ctrl:GetText()))) then ctrl:SetText(value) end
		end		
	end
	_ttip:CheckTip(tag, value)
end

local function _getSampleDetail(uf)
	local detail = {}
	local pdetail = Inspect.Unit.Detail("player")
	
	detail.calling = iif(uf.isplayer, pdetail.calling, "cleric")
	detail.charge = iif(uf.isplayer, pdetail.charge, nil)
	
	if (detail.calling == "mage") or (detail.calling == "cleric") then
		detail.mana = 800.5
		detail.manaMax = 1000
	elseif (detail.calling == "warrior") then
		detail.power = 40.5
	elseif (detail.calling == "rogue") then
		detail.energy = 60.5
		detail.energyMax = 100
	end
	
	detail.combat = true
	detail.faction = iif(uf.isplayer, pdetail.faction, iif(pdetail.faction == "defiant", "guardian", "defiant"))	
	detail.factionName = L[detail.faction]
	detail.guaranteedLoot = not (uf.isplayer or notnil(uf.group) or uf.ispet)
	detail.guild = "Some Guild"
	detail.health = 900.5
	detail.healthMax = 1000
	detail.healthCap = 1000
	detail.level = 50
	detail.locationName = "Somewhere"
	detail.mark = 2
	detail.name = iif(uf.isplayer, pdetail.name, "Someone")
	detail.planar = iif(uf.isplayer or uf.group, 2, nil)
	detail.planarMax = iif(uf.isplayer or uf.group, 4, nil)
	detail.player = true
	detail.pvp = true
	detail.race = "race"
	detail.raceName = "Race"
	detail.ready = true
	detail.relation = iif(uf.ispet or uf.isgroup or uf.isplayer, "friendly", "hostile")
	detail.role = iif(uf.isplayer, "heal", iif(uf.ispet, "dps", "support"))
	detail.tagged = iif(uf.isplayer or uf.group or uf.ispet, nil, true)
	detail.tier = iif(uf.isplayer or uf.group or uf.ispet, nil, "raid")
	detail.vitality = iif(uf.isplayer or uf.group, 80, nil)
	
	
	-- force some icons
	detail._showDead = true
	
	detail._isSample = true
	
	return detail
end
local function _updateUF(uf)
	local show = false
	
	if (uf.enabled) then	
		local detail = Inspect.Unit.Detail(uf.unit)
		local lastdetail = uf.lastDetail
		local reset = uf.reset
		
		
		-- doing configuration ?
		if (uf.mover:GetVisible()) then
			detail = _getSampleDetail(uf)
		else
			if (lastdetail) and (lastdetail._isSample) then
				reset = true
			end
		end
		
		if (reset) then
			lastdetail = {}
			-- clear out all text. don't worry about the full-path table lookup. This only executes when settings are changed in the config UI
			if (uf.titlebar) then uf.titlebar.text1:SetText(" "); uf.titlebar.text2:SetText(" ") end
			if (uf.healthbar) then uf.healthbar.text1:SetText(" "); uf.healthbar.text2:SetText(" ") end
			if (uf.powerbar) then uf.powerbar.text1:SetText(" "); uf.powerbar.text2:SetText(" ") end
			if (uf.chargebar) then uf.chargebar.text1:SetText(" "); uf.chargebar.text2:SetText(" ") end
			if (uf.extra) then uf.extra.text1:SetText(" "); uf.extra.text2:SetText(" ") end
			
			for _, icon in pairs(uf.icons) do
				icon:SetVisible(false)
			end
		end
		
		
		
		if (detail) then
			show = true
			
			local tip = _ttip._tip
			local indicators = uf.indicators
			local indicators2 = uf.indicators2
			local text_values = uf.text_values
			
			--if uf.isplayer then	-- this is a hack to fix the situation where some stats aren't updating after zoning.
				uf.updateCount = uf.updateCount + 1
				if (uf.updateCount > 20) or (detail.relation ~= lastdetail.relation) then
					uf.updateCount = 0
					lastdetail = {}
				end
			--end
			
			-- calling
			if (lastdetail.calling ~= detail.calling) or (tip == "calling") or (uf.reset) then				
				_applyIndicators(indicators.calling, indicators2.calling, detail.calling, colors.calling[detail.calling])					
				
				local sz; if (detail.calling) then sz = L[detail.calling] end
				_applyValues(text_values.calling, "calling", sz)
				
				if (uf.portrait:GetVisible()) then
					local texture = _aprofile.textures[ifnil(detail.calling, "unknown")]
					uf.portrait.texture:SetTexture(texture.src, texture.file)
				end								
			end
			
			-- relation
			if (lastdetail.relation ~= detail.relation) or (tip == "rel") then
				local relation = ifnil(detail.relation, "neutral")
				_applyIndicators(indicators.relation, indicators2.relation,relation, colors.relation[relation])
				_applyValues(text_values.rel, "rel", L[relation])
			end
			
			-- name
			if ((text_values.name) and (lastdetail.name ~= detail.name)) or (tip == "name") then
				local name = ifnil(detail.name, "??")				
				_applyValues(text_values.name, "name", name)
			end
			
			-- level
			if (lastdetail.level ~= detail.level) or (tip == "lvl") then			
				_applyValues(text_values.lvl, "lvl", tostring(detail.level))
			end
			
			
			-- planar
			if (lastdetail.planar ~= detail.planar) or (lastdetail.planarMax ~= detail.planarMax) or (tip == "planar") then
				local sz = ""
				if (detail.planar and detail.planarMax) then
					sz = tostring(detail.planar) .. "/" .. tostring(detail.planarMax)
				end
				_applyValues(text_values.planar, "planar", sz)				
			end
			
			-- raid mark
			if (lastdetail.mark ~= detail.mark) or (tip == "mark") then
				local sz = ""
				if (nil ~= detail.mark) then sz = tostring(detail.mark) end
				_applyValues(text_values.mark, "mark", sz)
			end
			
			--vitality
			if (lastdetail.vitality ~= detail.vitality) or (tip == "vit") then
				local sz
				if (detail.vitality) then sz = tostring(detail.vitality) .. "%" end
				_applyValues(text_values.vit, "vit", sz)
				
				_setIcon(uf, "vitality", iif((detail.vitality==nil) or (detail.vitality==100), nil, "vitality"))
			end
			
			-- offline / afk / warfront
			local status
			
			if (lastdetail.offline ~= detail.offline) and (detail.offline) then
				status = L["(offline)"]
			elseif (lastdetail.afk ~= detail.afk) and (detail.afk) then
				status = L["(afk)"]
			elseif (lastdetail.warfront ~= detail.warfront) and (detail.warfront) then
				status = L["(warfront)"]
			end
			
			-- health
			if (lastdetail.health ~= detail.health) or (lastdetail.healthMax ~= detail.healthMax) or (lastdetail.health == nil) or (lastdetail.healthMax == nil)
				or (tip == "hp") or (tip == "hp%") or (tip == "-hp") then
				local health = ceil(ifnil(detail.health,ifnil(detail.healthMax, 0))) -- in case we're in config mode
				
				local max = detail.healthMax
				local lowHealth = false
				local phealth = 0
				
				if (max == nil) or (health == nil) or (max == 0) then
					status = L["(unknown)"]
				else		
					phealth = health / max
					if (health <= 0) then 
						status = L["(dead)"]
					else
						if (uf.healthbar.value:GetAlpha() ~= 1) then uf.healthbar.value:SetAlpha(1) end
						
						
						lowHealth = (phealth < _aprofile.lowHealth)
						
						if max >= 1000000 then
							_applyValues(text_values.hp, "hp", string.format("%.3fM/%.3fM", health/1000000,max/1000000))
						else
							_applyValues(text_values.hp, "hp", tostring(health) .. " / " .. tostring(max))
						end
			
						_applyValues(text_values["hp%"], "hp%", string.format("%02d", phealth * 100) .. "%")
						_applyValues(text_values["-hp"], "-hp", iif(max == health, "", "-" .. tostring(max-health)))
						
					end					
				end
			
				detail._lowHealth = lowHealth
				if (lowHealth ~= lastdetail._lowHealth) then										
					_applyIndicators(indicators.lowHealth, indicators2.lowHealth,lowHealth, colors.misc_colors.lowHealth)					
				end
				
				
				local cxhealth = uf.healthbar.cx
				if (cxhealth == nil) or (cxhealth == 0) then cxhealth = uf.healthbar:GetWidth() end
				
				uf.healthbar.value:SetWidth(minmax(cxhealth * phealth, 0, cxhealth))
			end
			
			detail._status = status
			
			-- status
			if (status ~= lastdetail._status) then
				uf.healthbar.value:SetAlpha(iif(status==nil, 0.2, 1))
				if (status) then
					-- overrides health text values
					_applyValues(text_values.hp, "hp", status)
					_applyValues(text_values["hp%"], "hp%", status)
					_applyValues(text_values["-hp"], "-hp", status)					
				end
				
				_setIcon(uf, "dead", iif(status == L["dead"], "dead", nil))
			end
			
			if (detail._showDead) then		-- force for config
				--_setIcon(uf, "dead", "dead")
			end
			
			--power
			local power_type
			
			if (detail.power) then	
				power_type = "power"
			elseif (detail.mana) then
				power_type = "mana"
			elseif (detail.energy) then
				power_type = "energy"
			end
			
			detail._power_type = power_type	-- for next time
				
			if (power_type ~= lastdetail._power_type) or (tip == "pwr") or (tip == "pwr%") then
				uf.powerbar.value:SetBackgroundColor(rgba(ifnil(colors.bars[power_type], {0,0,0,0})))	
				if (not power_type) then
					_applyValues(text_values.pwr, "pwr", "")
					_applyValues(text_values["pwr%"], "pwr%", "")
				end
			end
			
			local cxpower = uf.powerbar.cx
			
			
			if detail.power and ((detail.power ~= lastdetail.power) or (tip == "pwr") or (tip == "pwr%")) then				
				detail.power = ceil(detail.power)
				
				_applyValues(text_values.pwr, "pwr", tostring(detail.power))
				_applyValues(text_values["pwr%"], "pwr%", tostring(detail.power))
				uf.powerbar.value:SetWidth(minmax(cxpower * (detail.power/100), 0, cxpower))
			elseif (detail.manaMax
				and ((detail.mana == nil) or (detail.mana ~= lastdetail.mana) or (detail.manaMax ~= lastdetail.manaMax) or (tip == "pwr") or (tip == "pwr%"))) then
				
				if (detail.mana == nil) then detail.mana = detail.manaMax end
				
				detail.mana = ceil(detail.mana)
				
				_applyValues(text_values.pwr, "pwr", tostring(detail.mana) .. " / " .. tostring(detail.manaMax))
				_applyValues(text_values["pwr%"], "pwr%", string.format("%02d", 100 * (detail.mana/detail.manaMax)) .. "%")
				
				local lowMana = ((detail.mana/detail.manaMax) < _aprofile.lowMana)
				detail._lowMana = lowMana
				
				if (lowMana ~= lastdetail._lowMana) then
					_applyIndicators(indicators.lowMana, indicators2.lowMana, lowMana, colors.misc_colors.lowMana)
				end
				uf.powerbar.value:SetWidth(minmax(cxpower * (detail.mana/detail.manaMax), 0, cxpower))
			elseif (detail.energyMax 
				and ((detail.energy ~= lastdetail.energy) or (detail.energyMax ~= lastdetail.energyMax) or (tip == "pwr") or (tip == "pwr%"))) then
				
				if (detail.energy == nil) then detail.energy = detail.energyMax end
				
				detail.energy = ceil(detail.energy)
				
				_applyValues(text_values.pwr, "pwr", tostring(detail.energy))
				_applyValues(text_values["pwr%"], "pwr%", string.format("%02d", 100 * (detail.energy/detail.energyMax)) .. "%")
				
				uf.powerbar.value:SetWidth(minmax(cxpower * (detail.energy/detail.energyMax), 0, cxpower))
			end
			
			-- charge
			if (uf.isplayer and (not uf.chargeShown) and (detail.calling == "mage"))then
				uf:Layout(_, true)
			end
			
			if (uf.isplayer and detail.charge and (detail.charge ~= lastdetail.charge)) or (uf.isplayer and tip=="chrg") then
				if (uf.chargebar) then
					local cxCharge = uf.chargebar.cx				
					uf.chargebar.value:SetWidth(minmax(cxCharge * (detail.charge / 100), 0, cxCharge))
				end
				_applyValues(text_values.chrg, "chrg", tostring(detail.charge))
			end
			
			-- guild
			if (lastdetail.guild ~= detail.guild) or (tip == "guild") then
				_applyValues(text_values.guild, "guild", detail.guild)
			end
			
			-- location
			if (lastdetail.locationName ~= detail.locationName) or (tip == "loc") then
				_applyValues(text_values.loc, "loc", detail.locationName)
			end
			
			-- race
			if (lastdetail.race ~= detail.race) or (tip == "race") then
				_applyValues(text_values.race, "race", detail.raceName)
			end
			
			-- faction
			if (lastdetail.faction ~= detail.faction) or (tip == "fac") then
				_applyValues(text_values.fac, "fac", detail.factionName)
				_setIcon(uf, "faction", detail.faction)
			end
			
			-- player
			if (lastdetail.player ~= detail.player) then
				_setIcon(uf, "player", iif(detail.player, "player", nil))
			end
			
			-- pvp
			if (lastdetail.pvp ~= detail.pvp) then
				_setIcon(uf, "pvp", iif(detail.pvp, "pvp", nil))
			end
			
			-- loot			
			if (lastdetail.guaranteedLoot ~= detail.guaranteedLoot) then
				_setIcon(uf, "loot", iif(detail.guaranteedLoot, "loot", nil))
			end
			
			-- ready check
			if (lastdetail.ready ~= detail.ready) then
				_setIcon(uf, "ready", iif(detail.ready, "ready", nil))
			end
			
			-- role
			if (lastdetail.role ~= detail.role) then
				local icon 
				if (detail.role == "tank") then icon = "tank" 
				elseif (detail.role == "heal") then icon = "heal"
				elseif (detail.role == "support") then icon = "support"
				elseif (detail.role == "dps") then icon = "dps"
				end
				
				_setIcon(uf, "role", icon)
			end
			
			-- LoS
			if (lastdetail.blocked ~= detail.blocked) then
				if (detail.blocked) then
					--uf:SetAlpha(0.4)--_aprofile.noLoS_Alpha)
					uf.los:SetVisible(true)
				else
					--uf:SetAlpha(1) --_aprofile.LoS_Alpha)
					uf.los:SetVisible(false)
				end
			else
				if (not detail.blocked) and uf.los:GetVisible() then uf.los:SetVisible(false) end
			end
			
			
			-- tier
			if (lastdetail.tier ~= detail.tier) then
				_setIcon(uf, "tier", detail.tier)
			end
			
			-- tagged
			if (lastdetail.tagged ~= detail.tagged) then
				local icon
				if (detail.tagged == "other") then
					icon = "tagged_other"
				elseif (detail.tagged == true) then
					icon = "tagged_true"
				end
				_setIcon(uf, "tagged", icon)
			end
			
			-- combat
			if  (indicators.combat) and (lastdetail.combat ~= detail.combat) then				
				_applyIndicators(indicators.combat, indicators2.combat, detail.combat, colors.misc_colors.combat)				
			end
			
			
			-- aggro
			if (indicators.aggro ~= detail.aggro) then
				_applyIndicators(indicators.aggro, indicators2.aggro, detail.aggro, colors.misc_colors.aggro)
			end	
			uf.lastDetail = detail
		else
			uf.lastDetail = {}
			for _, icon in pairs(uf.icons) do
				icon:SetVisible(false)
			end
			
			if (uf:GetVisible()) then
				for _, ctrl in ipairs(uf.text_values) do
					ctrl:SetText(" ")
				end
			end
		end
		
	end
	uf.reset = false
	
	if (uf:GetVisible() ~= show) then uf:SetVisible(show) end
	if (not show) then uf.los:SetVisible(false) end
end
local function _newUF()
	local uf = myUI:CreateFrame("Frame")
	local mover = myUI:CreateFrame("Frame")
	local sizer = myUI:CreateFrame("Texture", mover)	
	local clicker = myUI:CreateFrame("Frame")
	
	uf.mover = mover
	mover.sizer = sizer
	uf.clicker = clicker
	
	mover:SetSecureMode("normal")
	mover:SetLayer(LAYER_MOVER)
	mover:SetBackgroundColor(rgba(myuicolors.Mover))
	mover:SetVisible(false)
	
	mover.unittext = myUI:CreateFrame("Text", mover)
	mover.unittext:SetPoint("BOTTOMLEFT", mover, "TOPLEFT", 0, -4)
	mover.unittext:SetFontSize(12); mover.unittext.customFontSize = true	
	mover.unittext:SetBackgroundColor(0.6,0,0,0.6)
	
	mover.coordtext = myUI:CreateFrame("Text", mover)
	mover.coordtext:SetPoint("TOPLEFT", mover.unittext, "TOPRIGHT", 4, 0)
	mover.coordtext:SetFontSize(12); mover.coordtext.customFontSize = true
	mover.coordtext:SetBackgroundColor(0.6,0,0,0.6)
	
	mover.sizetext = myUI:CreateFrame("Text", mover)
	mover.sizetext:SetPoint("TOPLEFT", mover.coordtext, "TOPRIGHT", 4, 0)
	mover.sizetext:SetFontSize(12); mover.sizetext.customFontSize = true
	mover.sizetext:SetBackgroundColor(0.6,0,0,0.6)
	
	mover.sensitivity = myUI:CreateFrame("Text", mover)
	mover.sensitivity:SetPoint("TOPLEFT", mover.sizetext, "TOPRIGHT", 4, 0)
	mover.sensitivity:SetFontSize(12); mover.sensitivity.customFontSize = true
	mover.sensitivity:SetBackgroundColor(0.6,0,0,0.6)
	mover.sensitivity:SetText("+/- " .. tostring(_sensitivity))
	
	mover.x = 0
	mover.y = 0
	mover.dragging = false
	
	mover.Event.LeftDown = _leftDownM
	mover.Event.LeftUp = _leftUpM
	mover.Event.LeftUpoutside = _leftUpM
	mover.Event.MouseMove = _mouseMoveM
	mover.Event.RightDown = _rightDownM
	
	
	
	sizer:SetSize(18)
	sizer:SetTexture("myUI", "images/dragger.png")
	sizer:SetPoint("BOTTOMRIGHT", mover, "BOTTOMRIGHT", 0, 0)
	sizer.x = 0
	sizer.y = 0
	sizer.sizing = false
	
	sizer.Event.LeftDown = _leftDownS
	sizer.Event.LeftUp = _leftUpS
	sizer.Event.LeftUpoutside = _leftUpS
	sizer.Event.MouseMove = _mouseMoveS
	
	uf:SetSecureMode("normal")
	uf:SetAllPoints(mover)
	uf:SetLayer(LAYER_UF)
	
	clicker:SetSecureMode("restricted")
	clicker:SetAllPoints(mover)
	clicker:SetLayer(LAYER_UF - 1)
	
	uf.portrait = myUI:CreateFrame("Frame", uf)
	uf.portrait.texture = myUI:CreateFrame("Texture", uf.portrait)
	uf.portrait.texture:SetBackgroundColor(0,0,0,1)
	uf.portrait.texture:SetPoint("CENTER", uf.portrait, "CENTER")
	
	uf.titlebar = myUI:CreateFrame("Frame", uf)
	uf.titlebar:SetLayer(LAYER_BAR+1)
	uf.titlebar.text1 = myUI:CreateFrame("Text", uf.titlebar); uf.titlebar.text1.customFontSize = true; uf.titlebar.text1:SetLayer(LAYER_TEXT+1)
	uf.titlebar.text2 = myUI:CreateFrame("Text", uf.titlebar); uf.titlebar.text2.customFontSize = true; uf.titlebar.text2:SetLayer(LAYER_TEXT)
	uf.titlebar.value = myUI:CreateFrame("Frame", uf.titlebar); uf.titlebar.value:SetLayer(LAYER_BAR_VALUE+2)
	
	
	
	uf.healthbar = myUI:CreateFrame("Frame", uf)
	uf.healthbar:SetLayer(LAYER_BAR)	
	uf.healthbar.text1 = myUI:CreateFrame("Text", uf.healthbar); uf.healthbar.text1.customFontSize = true; uf.healthbar.text1:SetLayer(LAYER_TEXT)
	uf.healthbar.text2 = myUI:CreateFrame("Text", uf.healthbar); uf.healthbar.text2.customFontSize = true; uf.healthbar.text2:SetLayer(LAYER_TEXT+1)
	uf.healthbar.value = myUI:CreateFrame("Frame", uf.healthbar); --uf.healthbar.value:SetLayer(LAYER_BAR+3)
	
	uf.healthbar.texture = myUI:CreateFrame("Texture", uf.healthbar); uf.healthbar.texture:SetLayer(uf.healthbar.value:GetLayer()-1)
	uf.healthbar.texture:SetTexture("myUIUF", "images/smooth.png")
	uf.healthbar.texture:SetAllPoints(uf.healthbar.value)
	
	uf.powerbar = myUI:CreateFrame("Frame", uf)
	uf.powerbar:SetLayer(LAYER_BAR + 2)
	uf.powerbar.text1 = myUI:CreateFrame("Text", uf.powerbar); uf.powerbar.text1.customFontSize = true; uf.powerbar.text1:SetLayer(LAYER_TEXT)
	uf.powerbar.text2 = myUI:CreateFrame("Text", uf.powerbar); uf.powerbar.text2.customFontSize = true; uf.powerbar.text2:SetLayer(LAYER_TEXT+1)
	uf.powerbar.value = myUI:CreateFrame("Frame", uf.powerbar); --uf.powerbar.value:SetLayer(LAYER_BAR+1)
	
	uf.powerbar.texture = myUI:CreateFrame("Texture", uf.powerbar); uf.powerbar.texture:SetLayer(uf.powerbar.value:GetLayer()-1)
	uf.powerbar.texture:SetTexture("myUIUF", "images/smooth.png")
	uf.powerbar.texture:SetAllPoints(uf.powerbar.value)
	
	
	uf.extra = {
		text1 = myUI:CreateFrame("Text", uf),
		text2 = myUI:CreateFrame("Text", uf)
	}
	
	uf.extra.text2.customFontSize = true; uf.extra.text2:SetLayer(LAYER_TEXT)
	uf.extra.text1.customFontSize = true; uf.extra.text1:SetLayer(LAYER_TEXT+1)
	
	-- LoS overlay
	uf.los = myUI:CreateFrame("Frame", uf)
	uf.los:SetAllPoints()
	uf.los:SetLayer(LAYER_LOS)
	uf.los:SetBackgroundColor(0.1, 0.1, 0.1, 0.7)
	uf.los:SetVisible(false)
	
	-- etc ...
	uf.icons = {}
	
	
	uf.Initialize = _initUF
	uf.Dispose = _disposeUF
	uf.Layout = _layoutUF
	uf.Update = _updateUF
	uf.OnCombat = _onCombatUF
	uf.Reset = _resetUF
	
	return uf
end




local function _recycleUnitFrame(uf)
	uf:Dispose()
	tinsert(_recycledUnitFrames, uf)
end
local function _getUnitFrame(unit, group)
	local uf 
	
	if (#_recycledUnitFrames > 0) then
		uf = tremove(_recycledUnitFrames)		
	else
		uf = _newUF()
	end
	
	uf:Initialize(unit, group)
	uf.enabled = true
	return uf	
end




local function _setupTooltip()
	if (not _ttip) then
		-- create our tooltip
		
		_ttip = myUI:CreateFrame("Frame")
		_ttip.text = myUI:CreateFrame("Text", _ttip)	
		
		_ttip.SetTip = function(self, tip)
			if (tip == "nil") then tip = nil end
			
			if (tip ~= self._tip) then
				self._tip = tip
				
				if (tip == nil) then 
					self:SetVisible(false) 
					self.text:SetText("")
				end
			end
			
		end
		
		_ttip.CheckTip = function(self, tip, value)
			if (self._tip) and (self._tip == tip) then
				-- this might apply to us
				if (isempty(value)) then
					self:SetVisible(false)
				else
					if (value ~= (self.text:GetText())) then
						self.text:SetText(tostring(value))
					end
					
					self:SetVisible(true)
				end
			elseif (self._tip == nil) then
				self:SetVisible(false)
				self.text:SetText("")
			end			
		end
		
		--_ttip.text:SetFontSize(10); _ttip.text.customFontSize = true
		_ttip:SetLayer(LAYER_TOOLTIP)
		
		local margin = 1.5
		_ttip:SetPoint("TOPLEFT", _ttip.text, "TOPLEFT", -margin, -margin)
		_ttip:SetPoint("BOTTOMRIGHT", _ttip.text, "BOTTOMRIGHT", margin, margin)		
	end
	
	_ttip.text:SetFontColor(rgba(colors.ui.ttipfg))
	_ttip:SetBackgroundColor(rgba(colors.ui.ttipfg))
	_ttip.text:SetBackgroundColor(rgba(colors.ui.ttipbg))	
end

local function _resetAll()
	for _, frame in pairs(_frames) do
		frame:Reset()
	end
end

local function _layoutAll(doExternal, barsOnly)
	for _, frame in pairs(_frames) do
		frame:Layout(doExternal, barsOnly)
	end
	
	_setupTooltip()
end


local function _enableUnit(unit, enabled)
	_aprofile.frames[unit].enabled = enabled

	local f
	
	if (sfind(unit, "group")) then
		local gunit 
		for i=1,20 do
			gunit = gsub(unit, "%d%d", lpad(tostring(i), 2, "0"))
			f = _frames[gunit]
			if (f) then
				f.enabled = enabled
				if (not enabled) then
					--_recycleUnitFrame(f)
					f.mover:SetVisible(false)
				else
					f.mover:SetVisible(true)
				end
			elseif (enabled) then
				_frames[gunit] = _getUnitFrame(gunit, 1)
			end
		end	

		-- we need to do this for group frames b/c not all group frames existed for the initilialize call when new ones were created
		for i=1,20 do
			gunit = gsub(unit, "%d%d", lpad(tostring(i), 2, "0"))
			f = _frames[gunit]
			if (f and f.enabled) then
				f:Layout(true)
			end
		end		
	else
		f = _frames[unit]
		if (f) then
			f.enabled = enabled
			if (not enabled) then
				--_recycleUnitFrame(f)
				f.mover:SetVisible(false)
			else
				f.mover:SetVisible(true)
			end
		elseif (enabled) then
			_frames[unit] = _getUnitFrame(unit)
		end
		
		if (f and enabled) then f:Layout(true) end
	end	
end




-- public interface methods

function unitframes:Initialize()
	local id = unitframes.identifier
	
	table.insert(Event.System.Secure.Enter, {_secureEnter, id, "_secureEnter"})
	table.insert(Event.System.Secure.Leave, {_secureLeave, id, "_secureLeave"})
	
	--[[	Unit update notes:
	
		The unit.changed event model has some real problems. It's too granular and waaaaay too spammy during high activity times. Additionally,
		it doesn't appear to be throttled, so that during high activity times, addon CPU usage will spike and seriously lag both the addon and
		the Rift client itself. Furthermore, each particular detail change still requires a full Inspect.Unit.Detail call and subsequent redunant
		evaluation of that detail.
		
		Instead, I'm using a simple update timer. This lets me balance CPU load vs. responsiveness. CPU usage is slightly higher during low
		activity times, but that usually isn't an issue unless a machine is severely under-powered. During high activity times, this lets me
		cap the addon's CPU usage at a configurable level. The default is 10 times/second which is still faster than a human can see
		and comprehend the change in information.
	]]--
	myUI.Events.UnitUpdate:Subscribe(id .. "._unitUpdate", self, _unitUpdate)
	
end

function unitframes:LoadProfile(account, character)
	_aprofile = account
	_cprofile = character

	
	if (not _aprofile.runonce) then
		local adefs = unitframes._getAccountDefaults()
	
		myUI.Utility.nilCopy(_aprofile, adefs)
		_aprofile.runonce = true
	end
	
	if (_aprofile.colors.misc_colors.aggro == nil) then	-- new setting added
		_aprofile.colors.misc_colors.aggro = {1,0,0,1}
	end
	
	if (not _aprofile.icons.ready) then
		-- whoops
		_aprofile.icons.ready = {anchor = "CENTER", cxy = 32, show = false, tooltip = "nil", xoff = 0, yoff = 0}
	end
	
	for _, frame in pairs(_frames) do 
		_recycleUnitFrame(frame)		
	end
	_frames = {}
	
	for unit, def in pairs(_aprofile.frames) do
		if (def.enabled) then
			if (sfind(unit, "01")) then
				local gunit
				
				for i=1,20 do
					gunit = gsub(unit, "%d%d", lpad(tostring(i), 2, "0"))
					_frames[gunit] = _getUnitFrame(gunit, i)				
				end
			else
				_frames[unit] = _getUnitFrame(unit)
			end
		end
	end
	
	colors = _aprofile.colors
	
	_setupTooltip()
	
	if (_configFrame) then _configFrame:LoadProfile() end	
end

function unitframes:GetConfigurationFrame(parent)
	if (not _configFrame) then
		Command.System.Watchdog.Quiet()
		_configFrame = myUI:CreateFrame("Frame", parent)	
		
		local margin = 16
		local tstrip = myUI:CreateFrame("myUI_Tabstrip", _configFrame)
		
		tstrip:SetPoint("TOPLEFT", _configFrame, "TOPLEFT")
		tstrip:SetPoint("BOTTOMRIGHT", _configFrame, "BOTTOMLEFT", 160, 0)
		
		local tpages = {}
		local clrDropdownBg = {0.1, 0.1, 0.1, 1}
		
		local function _loadUnit(unit)
			_configFrame.activeUnit = nil -- signifies that we're loading for controls that throw events when setting properties in code
			
			
			local st = _configFrame.TextValuePicker:GetItem(1)
			local ind = _configFrame.IndicatorPicker:GetItem(1)
			
			_configFrame.IndicatorPicker.target = nil
			_configFrame.IndicatorPicker:SetVisible(false)
			
			_configFrame.BgColorPicker.target = nil
			_configFrame.BgColorPicker:SetVisible(false)
			
			_configFrame.NumberPicker.target =nil
			_configFrame.NumberPicker:SetVisible(false)
			
			_configFrame.TextValuePicker.target = nil
			_configFrame.TextValuePicker:SetVisible(false)
			
			local config1
			local config2 = _aprofile
			local incombat = Inspect.System.Secure()
			
			local isdefaults = (unit == "defaults")
			local ismouseover = string.find(unit, "mouseover")
			
			
			if (isdefaults) then				
				config1 = config2	-- so our fallback is the same as the first on (ie effectively not there)
				
				if (st and st.value == nil) then -- ie, use none. Should not appear for the default settings. That should be 'nil' to not nil.
					_configFrame.TextValuePicker:RemoveItemAt(1)
				end
				if (ind and ind.value == nil) then
					_configFrame.IndicatorPicker:RemoveItemAt(1)
				end
			else
				config1 = _aprofile.frames[unit]				
				
				if (st and st.value ~= nil) then
					_configFrame.TextValuePicker:InsertItem(1, {value = nil, tostring = function(self) return L["use default"] end})
				end
				if (ind and ind.value ~= nil) then
					_configFrame.IndicatorPicker:InsertItem(1, {value = nil, tostring = function(self) return L["use default"] end})
				end
			end
			
			local barControls = _configFrame.FrameBarControls
			
			-- load bar settings
			barControls.portrait.show:SetChecked(ifnil(config1.portrait.show, config2.portrait.show))
			
			for bar, ctrls in pairs(barControls) do
				if (ctrls.show) then
					ctrls.show:SetChecked(ifnil(config1[bar].show, config2[bar].show))
				end
				if (ctrls.margin) then
					ctrls.margin:SetText(tostring(ifnil(config1[bar].margin, config2[bar].margin)))
				end
				if (ctrls.bgcolor) then
					ctrls.bgcolor:SetBackgroundColor(rgba(ifnil(config1[bar].bgcolor, config2[bar].bgcolor)))
				end
				
				if (ctrls.fgcolor) then					
					ctrls.fgcolor:SetBackgroundColor(rgba(ifnil(config1[bar].fgcolor, config2[bar].fgcolor)))
				end
				
				if (ctrls.fontsize) then					
					ctrls.fontsize:SetText(tostring(ifnil(config1[bar].fontsize, config2[bar].fontsize)))
				end
				if (ctrls.text) then
					local text = config1[bar].text
					if (isdefaults) then text = nilstr(text) end
					if (text == nil) then
						ctrls.text:SetText(L["use default"])
					elseif (text == "nil") then
						ctrls.text:SetText(L["use empty"])
					else
						ctrls.text:SetText(L[text])
					end
				end				
				--[[ 
					Indicator gets tricky
					if global setting, then "nil" means use nothing L["use empty"]
					if specific unit setting, then "nil" means use nothing regardless of default, while
						nil value means use default
					
				
				]]--
				
				if (ctrls.indicator) then
					local indicator = config1[bar].indicator
					if (isdefaults) then indicator = nilstr(indicator) end
					if (indicator == nil) then
						ctrls.indicator:SetText(L["use default"])
					elseif (indicator == "nil") then
						ctrls.indicator:SetText(L["use empty"])
					else
						ctrls.indicator:SetText(L[indicator])
					end
				end
				
				if (ctrls.indicator2) then
					local indicator = config1[bar].indicator2
					if (isdefaults) then indicator = nilstr(indicator) end
					if (indicator == nil) then
						ctrls.indicator2:SetText(L["use default"])
					elseif (indicator == "nil") then
						ctrls.indicator2:SetText(L["use empty"])
					else
						ctrls.indicator2:SetText(L[indicator])
					end
				end
				
				
				if (ctrls.cyratio) then
					ctrls.cyratio:SetPosition(ifnil(config1[bar].cyration, config2[bar].cyratio))
				end
				if (ctrls.align) then
					local align = ifnil(config1[bar].align, config2[bar].align)
					align = iif(align == "RIGHT", 2, 1)
					ctrls.align:SetPosition(align)
				end
				
				if (ctrls.xoff) then
					ctrls.xoff:SetText(tostring(ifnil(config1[bar].xoff, config2[bar].xoff)))
				end
				if (ctrls.yoff) then
					ctrls.yoff:SetText(tostring(ifnil(config1[bar].yoff, config2[bar].yoff)))
				end
				
				if (ctrls.reset) then
					ctrls.reset:SetEnabled(not isdefaults)
					ctrls.reset:SetVisible(not isdefaults)
				end
				
			
				
				if (bar == "icons") then	
				
					local icons1 = config1.icons
					local icons2 = config2.icons
				
					if (icons2 == nil) then icons2 = {} end
					if (icons1 == nil) then icons1 = icons2 end
					
				
					for icon, icon_ctrls in pairs(ctrls) do
						if (icon_ctrls) then
							if (icons2[icon] == nil) then
								--[[  2-10-12					
									Discovered a tricky little bug. If a particular unit frame has never been enabled for a non-default
									profile, 'icons2' above will be nil which causes an error.
								]]--		
								icons2[icon] = {
									anchor = "LEFTCENTER",
									cxy = 28,
									show = false,
									tooltip = "nil",
									xoff = 0,
									yoff = 0
								}
								
							end
							if (icons1[icon] == nil) then icons1[icon] = {} end
						
							if (icon_ctrls.show) then								
								icon_ctrls.show:SetChecked(ifnil(icons1[icon].show, icons2[icon].show))
							end
							if (icon_ctrls.cxy) then
								icon_ctrls.cxy:SetText(tostring(ifnil(icons1[icon].cxy, icons2[icon].cxy)))
							end
							if (icon_ctrls.xoff) then
								icon_ctrls.xoff:SetText(tostring(ifnil(icons1[icon].xoff, icons2[icon].yoff)))
							end
							if (icon_ctrls.yoff) then
								icon_ctrls.yoff:SetText(tostring(ifnil(icons1[icon].yoff, icons2[icon].yoff)))
							end
							if (icon_ctrls.anchor) then
								icon_ctrls.anchor:SetText(ifnil(icons1[icon].anchor, icons2[icon].anchor))
							end
							if (icon_ctrls.tooltip) then
								local ttip = icons1[icon].tooltip
								if (isdefaults) then ttip = nilstr(ttip) end
								if (ttip == nil) then
									ttip = L["use default"]
								elseif (ttip == "nil") then
									ttip = L["use empty"]
								else
									ttip = L[ttip]
								end
								
								icon_ctrls.tooltip:SetText(ttip)
							end
							if (icon_ctrls.reset) then
								icon_ctrls.reset:SetEnabled(not isdefaults)
								icon_ctrls.reset:SetVisible(not isdefaults)
							end
						end
					end
				end
				
				if (bar == "actions") then
					local actions1 = config1.actions
					local actions2 = config2.actions
					
					for action, macro_controls in pairs(ctrls) do
						
						if (ismouseover) then
							macro_controls.macro:SetVisible(false)
						else
							local m = actions1[action]
							
							if (m == nil) and (not isdefaults) then
								m = L["default action"]
							else
								m = ifnil(m, "")
							end
							
							macro_controls.macro:SetText(m)
							macro_controls.macro:SetVisible(true)
						end
						macro_controls.apply:SetEnabled(not (incombat or ismouseover))
						macro_controls.apply:SetVisible(not ismouseover)
						
						macro_controls.reset:SetVisible(not (isdefaults or incombat or ismouseover))
						macro_controls.reset:SetEnabled(not (isdefaults or incombat or ismouseover))
					end
					
				end
			end
			
			_configFrame.activeUnit = unit
		end
		
		if (true) then	-- globals tab page
			local tpage = myUI:CreateFrame("myUI_Tabpage", _configFrame)
			tpage:SetPoint("TOPLEFT", tstrip, "TOPRIGHT")
			tpage:SetPoint("BOTTOMRIGHT", _configFrame, "BOTTOMRIGHT")
			tpages["global"] = tpage
			
			tstrip:AddTab(L["Global"])
			
			local vpbars = myUI:CreateFrame("myUI_Viewport", tpage)
			vpbars:SetPoint("TOPLEFT", tpage, "TOPLEFT", margin, margin)
			vpbars:SetPoint("BOTTOMRIGHT", tpage, "BOTTOMRIGHT", -margin/2, -margin)
				
			local content = vpbars.Content
				
			content:SetHeight(700)
			--content:SetBackgroundColor(1,0,0,0.1)  -- just to see how big it is when testing
			
			if (true) then	-- enabled units				
				local label = myUI:CreateFrame("Text",content)
				label:SetText(L["Enabled Units"])
				label:SetBackgroundColor(rgba(myuicolors.Hover))
				label:SetPoint("TOPLEFT", content, "TOPLEFT")
				label:SetPoint("TOPRIGHT", content, "TOPRIGHT")
				
				tpage._unitsBox = bbox				
				_configFrame.UnitChecks = {}
				
				local function _onUnitChecked(checkbox)
					if (not _configFrame._loading) then
						if (not Inspect.System.Secure()) then
							local unit = checkbox.unit
							
							_enableUnit(unit, checkbox:GetChecked())
						end
					end
				end
				
				
				local lastCheck
				for i, u in ipairs(_available_units) do
					local chk = myUI:CreateFrame("myUI_Checkbox",content)
					chk:SetText(L["u-" .. u])
					chk.unit = u
					chk.Event.CheckboxChange = _onUnitChecked
					
					_configFrame.UnitChecks[u] = chk
					lastCheck = chk
					
					if (i == 1) then
						chk:SetPoint("TOPLEFT", label, "BOTTOMLEFT", 4, 4)
					elseif (i == 5) or (i == 9) then
						chk:SetPoint("TOPLEFT", _configFrame.UnitChecks[_available_units[i-4]], "BOTTOMLEFT", 0, 8)
					else
						chk:SetPoint("LEFTCENTER", _configFrame.UnitChecks[_available_units[i-1]], "RIGHTCENTER", 140, 0)
					end
				end		
			end
			
			local lastbox
			if (true) then 	-- colors	
				_configFrame.ColorBoxes = {}
			
				local offy = 20 + _configFrame.UnitChecks["mouseover.target"]:GetBottom() - content:GetTop()
				local label = myUI:CreateFrame("Text", tpage)
				label:SetText(L["Indicator Colors"])
				label:SetPoint("TOPLEFT", content, "TOPLEFT", 0, offy)
				label:SetPoint("TOPRIGHT", content, "TOPRIGHT", 0, offy)
				label:SetBackgroundColor(rgba(myuicolors.Hover))
				
				local function _colorClick(box)
					local cpicker = _configFrame.ColorPicker
					
					if (cpicker.box == box) and (cpicker:GetVisible()) then
						cpicker:SetVisible(false)
					else
						cpicker.box = box
						cpicker:SetPoint("TOPLEFT", box, "BOTTOMLEFT", 0, 2)
						cpicker:SetColor(box:GetBackgroundColor())
						cpicker:SetVisible(true)
					
					end
					
				end
				
				local function _colorPickerColorChanged(cpicker, r, g, b, a)					
					local box = cpicker.box
					box:SetBackgroundColor(r,g,b,a)
					
					_aprofile.colors[box.indicator][box.value] = {round(r,3), round(g,3), round(b, 3), round(a,3)}
					_layoutAll(false)
					_resetAll()
				end
				
				
				local picker = myUI:CreateFrame("myUI_ColorPicker", _configFrame)
				picker:SetVisible(false)
				picker:SetBackgroundColor(rgba(clrDropdownBg))
				picker.Event.ColorChanged = _colorPickerColorChanged
				
				_configFrame.ColorPicker = picker
				tpage._colorsBox = bbox
				
				local anchor = label --content
				local offx = 4
				local last
				for k, v in pairs(_aprofile.colors) do					
					label = myUI:CreateFrame("Text", content)
					label:SetText(L[k] .. ":")
				
					label:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", offx, 4)
					
					local x = 80
					for k2, v2 in pairs(v) do
						local box = myUI:CreateFrame("Frame", content)
						box:SetSize(16)
						box:SetPoint("LEFTCENTER", label, "LEFTCENTER", x, 0)
						x = x + 110
						
						box.indicator = k
						box.value = k2
						box.Event.LeftDown = _colorClick
						
						local label2 = myUI:CreateFrame("Text", content)
						label2:SetText(L[k2])
						label2:SetPoint("LEFTCENTER", box, "RIGHTCENTER", 4, 0)
						
						table.insert(_configFrame.ColorBoxes, box)
						
						last = label2
						lastbox = box
					end
					
					anchor = label				
					offx = 0
				end
				
				
			end
			
			if (true) then	--icons								
				local offy = 20 + lastbox:GetBottom() - content:GetTop()
				local label = myUI:CreateFrame("Text", tpage)
				label:SetText(L["icons"])
				label:SetPoint("TOPLEFT", content, "TOPLEFT", 0, offy)
				label:SetPoint("TOPRIGHT", content, "TOPRIGHT", 0, offy)
				label:SetBackgroundColor(rgba(myuicolors.Hover))
				
				local icons = {
					"tank", "heal", "dps", "support", "player",
					"cleric", "rogue", "warrior", "mage", "unknown",
					"pvp", "vitality", "ready", "loot", "dead", "group", "raid",
					"defiant", "guardian", "tagged_true", "tagged_other",
					
				}
				
				local offx = 4
				local offy = 28
				local anchor = label
				local icons2 = {}
				local last
				for i, setting in ipairs(icons) do
					local icon = myUI:CreateFrame("Texture", content)
					icon:SetSize(32)
					table.insert(icons2, icon)
					
					icon:SetTexture(_aprofile.textures[setting].src, _aprofile.textures[setting].file)
					icon:SetPoint("TOPLEFT", anchor, "TOPLEFT", offx, offy)
					
					icon.setting = setting
				
				
					local lbl = myUI:CreateFrame("Text", content)
					lbl:SetText(L[setting])
					lbl:SetPoint("LEFTCENTER", icon, "RIGHTCENTER", 4, 0)
				
					local row, col = math.modf(i/5)
					if (i > 4) and (col == 0) then							
						anchor = icons2[i-4]
						offx = 0
						offy = 38						
					else
						anchor = icon
						offx = 130
						offy = 0
					end
					
					last = icon
				end
				
				_configFrame._iconsBox = bbox
				
			end			
		end
		
		if (true) then 	-- units tab page		
			local tpage = myUI:CreateFrame("myUI_Tabpage", _configFrame)
			tpage:SetPoint("TOPLEFT", tstrip, "TOPRIGHT")
			tpage:SetPoint("BOTTOMRIGHT", _configFrame, "BOTTOMRIGHT")
			tpage:SetVisible(false)
			tpages["unit"] = tpage
			
			tstrip:AddTab({unit = "defaults", text = L["u-defaults"], tostring = function(t) return L["u-defaults"] end})
			for _, u in ipairs(_available_units) do
				tstrip:AddTab({unit = u, text = L["u-" .. u], tostring = function(t) return L["u-" .. u] end})				
			end
			
			
			if (true) then -- dropdowns
				if (true) then 	--  text-value dropdown
					local tvals = {
						"hp", "hp%", "-hp", "pwr", "pwr%", "chrg", "lvl", "name", "planar", "vit", "mark", "guild", "loc", "race", "fac", "calling", "rel"
					}
					local tpicker = myUI:CreateFrame("myUI_Listbox", tpage)
					tpicker:SetSize(120,160)
					
					tpicker:AddItem({value="nil", tostring = function(self) return L["use empty"] end})
					for _, v in ipairs(tvals) do
						tpicker:AddItem({value = v, tostring = function(self) return L[v] end})
					end
					
					tpicker.Content:SetBackgroundColor(rgba(clrDropdownBg))
					tpicker:SetSelectedColor(0,0,1,1)
					tpicker:SetLayer(10)
					tpicker:SetVisible(false)
					_configFrame.TextValuePicker = tpicker
				end
				
				if (true) then 	-- anchor-picker dropdown
					local abox = myUI:CreateFrame("myUI_Anchorbox", tpage)
					abox:SetSize(160,100)
					abox:SetBackgroundColor(rgba(clrDropdownBg))
					abox:SetLayer(10)
					abox:SetVisible(false)
					_configFrame.AnchorPicker = abox
				end
				
				if (true) then	-- color-indicator dropdown					
					local tvals = {
						"aggro", "relation", "calling", "tier", "role", "tagged", "lowHealth", "lowMana", "combat",
					}
					local tpicker = myUI:CreateFrame("myUI_Listbox", tpage)
					tpicker:SetSize(120,160)
					
					tpicker:AddItem({value="nil", tostring = function(self) return L["use empty"] end})
					for _, v in ipairs(tvals) do
						tpicker:AddItem({value = v, tostring = function(self) return L[v] end})
					end
					
					tpicker.Content:SetBackgroundColor(rgba(clrDropdownBg))
					tpicker:SetSelectedColor(0,0,1,1)
					tpicker:SetVisible(false)
					tpicker:SetLayer(10)
					_configFrame.IndicatorPicker = tpicker
				end
			
				if (true) then	-- color-picker for units
					local picker = myUI:CreateFrame("myUI_ColorPicker", _configFrame)
					picker:SetLayer(10)
					picker:SetVisible(false)
					picker:SetBackgroundColor(rgba(clrDropdownBg))					
					_configFrame.BgColorPicker = picker
				end
				if (true) then 	-- number/size slider picker
					local picker = myUI:CreateFrame("myUI_BorderBox", _configFrame)
					picker:SetVisible(false)
					picker:SetBackgroundColor(rgba(clrDropdownBg))
					local slider = myUI:CreateFrame("RiftSlider", picker)
					
					slider:SetPoint("LEFTCENTER", picker, "LEFTCENTER", 16, 8)
					slider:SetPoint("RIGHTCENTER", picker, "RIGHTCENTER", -16, 8)
					
					picker:SetSize(180, 32)
					picker:SetLayer(10)
					
					picker.slider = slider
					_configFrame.NumberPicker = picker
				end
			end
			
			
			
			local function getUnitSettings(unit)
				if (unit == "defaults") then
					return _aprofile
				else
					return _aprofile.frames[unit]
				end
			end
			
			local function layoutFrames(unit, doExternal, barsOnly)
				if (unit == "defaults") then
					for _, uf in pairs(_frames) do
						uf:Layout(doExternal, barsOnly)
						uf:Reset()
					end
				else
					if (string.find(unit, "group")) then
						local gunit 
						for i=1,20 do
							gunit = string.gsub(unit, "%d%d", lpad(tostring(i), 2, "0"))							
							if (_frames[gunit]) then
								_frames[gunit]:Layout(doExternal, barsOnly)
								_frames[gunit]:Reset()
							end
						end	
					else
						if (_frames[unit]) then
							_frames[unit]:Layout(doExternal, barsOnly)
							_frames[unit]:Reset()
						end
					end
				end
			end
			local function updateMacros(unit)
				if (unit == "defaults") then
					for _, uf in pairs(_frames) do
						_updateMacrosUF(uf)
					end
				else
					if (string.find(unit, "group")) then
						local gunit
						for i=1,20 do
							gunit = string.gsub(unit, "%d%d", lpad(tostring(i), 2, "0"))
							if (_frames[gunit]) then
								_updateMacrosUF(_frames[gunit])
							end
						end
					else
						if (_frames[unit]) then
							_updateMacrosUF(_frames[unit])
						end
					end
				end
			end
			
			-- event handlers for unit settings controls
			
			-- unit show checked 
			local function showChecked(chk)
				local unit = _configFrame.activeUnit
				if (unit) then
					-- get config section
					local settings = getUnitSettings(unit)[chk.item]
					
					settings.show = chk:GetChecked()
					layoutFrames(unit, false)					
				end
			end
			local function showIconChecked(chk)
				local unit = _configFrame.activeUnit
				if (unit) then
					local settings = getUnitSettings(unit).icons[chk.item]
					
					settings.show = chk:GetChecked()
					layoutFrames(unit, false)					
				end
			end
			
			-- indicator dropdown selected item changed
			local function indChanged(txt, selindexes)
				if (txt.target) then
					if (selindexes and (#selindexes > 0)) then
						local unit = _configFrame.activeUnit
						if (unit) then
							local v = txt:GetItem(selindexes[1])
							txt.target:SetText(v.tostring())
							
							txt:ClearSelections(true)
							txt:SetVisible(false)
							
							-- get the correct config section 
							local settings
							if (txt.target.section) then
								settings = getUnitSettings(unit)[txt.target.section][txt.target.item]
							else
								settings = getUnitSettings(unit)[txt.target.item]
							end
							local set = ifnil(txt.target.setting, "indicator") -- default
							settings[set] = v.value
							layoutFrames(unit, false)							
						end
					end
				end
			end
			
			
			-- text-values dropdown item selected
			local function textValueChanged(txt, selindexes)
				if (txt.target) then
					if (selindexes and (#selindexes > 0)) then
						local unit = _configFrame.activeUnit
						if (unit) then
							local v = txt:GetItem(selindexes[1])
							txt.target:SetText(v.tostring())
							
							txt:ClearSelections(true)
							txt:SetVisible(false)
							
							local settings 
							if (txt.target.section) then
								settings = getUnitSettings(unit)[txt.target.section][txt.target.item]
							else
								settings = getUnitSettings(unit)[txt.target.item]
							end
							local set = ifnil(txt.target.setting, "text")	-- default
							settings[set] = v.value
							layoutFrames(unit, false)				
						end
					end
				end
			end
			
			-- bars indicator box clicked
			local function indClick(txt)
				local ddown = _configFrame.IndicatorPicker
				
				_configFrame.NumberPicker:SetVisible(false)
				_configFrame.AnchorPicker:SetVisible(false)
				_configFrame.TextValuePicker:SetVisible(false)
				_configFrame.BgColorPicker:SetVisible(false)
				
				if (ddown.target == txt) then
					ddown:SetVisible(not ddown:GetVisible())
				else
					ddown.target = txt						
					ddown:ClearSelections(true)
					ddown:SetPoint("TOPLEFT", txt, "BOTTOMLEFT", 0, 2)
					ddown:SetVisible(true)						
				end
			end
			
			-- text value box clicked
			local function textClick(txt)
				local ddown = _configFrame.TextValuePicker
				
				_configFrame.IndicatorPicker:SetVisible(false)
				_configFrame.NumberPicker:SetVisible(false)
				_configFrame.AnchorPicker:SetVisible(false)
				_configFrame.BgColorPicker:SetVisible(false)
				
				if (ddown.target == txt) then
					ddown:SetVisible(not ddown:GetVisible())
				else
					ddown.target = txt						
					ddown:ClearSelections(true)
					ddown:SetPoint("TOPLEFT", txt, "BOTTOMLEFT", 0, 2)
					ddown:SetVisible(true)						
				end
			end
			-- anchor box clicked
			local function anchorClick(txt)
				local ddown = _configFrame.AnchorPicker
				
				_configFrame.IndicatorPicker:SetVisible(false)
				_configFrame.NumberPicker:SetVisible(false)				
				_configFrame.TextValuePicker:SetVisible(false)
				_configFrame.BgColorPicker:SetVisible(false)
				
				if (ddown.target == txt) then
					ddown:SetVisible(not ddown:GetVisible())
				else
					ddown.target = txt
					ddown:ClearAnchors()
					ddown:SetPoint("TOPLEFT", txt, "BOTTOMLEFT", 0, 2)
					ddown:SetVisible(true)
				end
			end
			
			-- bars color box clicked
			local function colorClick(clrbox)
				local cpicker = _configFrame.BgColorPicker
				
				_configFrame.IndicatorPicker:SetVisible(false)
				_configFrame.NumberPicker:SetVisible(false)
				_configFrame.AnchorPicker:SetVisible(false)
				_configFrame.TextValuePicker:SetVisible(false)
				
				if (cpicker.target == clrbox) then
					cpicker:SetVisible(not cpicker:GetVisible())
				else
					cpicker.target =clrbox
					cpicker:SetColor(clrbox:GetBackgroundColor())
					cpicker:SetPoint("TOPLEFT", clrbox, "BOTTOMLEFT", 0, 2)
					cpicker:SetVisible(true)
				end					
			end
			
			-- bgcolor dropdown color change
			local function bgcolorChanged(cpicker, r, g, b, a)					
				cpicker.target:SetBackgroundColor(r,g,b,a)
				
				local unit = _configFrame.activeUnit
				if (unit) then
					local settings = getUnitSettings(unit)[cpicker.target.item]
					local r, g, b, a = cpicker:GetColor()
					
					-- note, there are no items with both fgcolor and bgcolor but setting the other for any of them has no effect
					settings.bgcolor = {r,g,b,a}
					settings.fgcolor = {r,g,b,a}
				
					layoutFrames(unit, false)
				end
			end
		
			
		
			local function numberClick(box)
				local picker = _configFrame.NumberPicker
				
				_configFrame.IndicatorPicker:SetVisible(false)
				_configFrame.AnchorPicker:SetVisible(false)
				_configFrame.TextValuePicker:SetVisible(false)
				_configFrame.BgColorPicker:SetVisible(false)
				
				if (picker.target == box) then
					picker:SetVisible(not picker:GetVisible())
				else
					local holdUnit = _configFrame.activeUnit
					_configFrame.activeUnit = nil	-- needed to postion changed doesn't cause in infinite loop
					
					local range = box.max - box.min
					picker:SetWidth(minmax(range*2, 180, 800))
					
					picker.target = box
					picker.slider:SetRange(box.min, box.max)
					
					picker.slider:SetPosition(tonumber((box:GetText())))
					
					picker:SetPoint("TOPLEFT", box, "BOTTOMLEFT", 0, 2)
					picker:SetVisible(true)
					
					_configFrame.activeUnit = holdUnit
				end
			end
		
			-- margin slider changed
			local function marginChanged(slider)
				local unit = _configFrame.activeUnit
				if (unit) then
					local settings = getUnitSettings(unit)[slider.item]
					settings.margin = slider:GetPosition()
					
					layoutFrames(unit, false)
				
				end
			end
			-- cyratio changed (relative bar heights)
			local function cyratioChanged(slider)
				local unit = _configFrame.activeUnit
				if (unit) then
					local settings = getUnitSettings(unit)[slider.item]
					settings.cyratio = slider:GetPosition()
					
					layoutFrames(unit, false, true)				
				end
			end
			-- portrait alignment changed
			local function alignChanged(slider)
				local unit = _configFrame.activeUnit
				if (unit) then
					local settings = getUnitSettings(unit)[slider.item]
					settings.align = iif(slider:GetPosition() == 2, "RIGHT", "LEFT")
					
					layoutFrames(unit, false)
				end
			end
			-- number picker changed
			local function numberChanged(slider)
				local picker = slider:GetParent()
				local unit = _configFrame.activeUnit
				if (unit) then
					local settings
					if (picker.target.section) then
						settings = getUnitSettings(unit)[picker.target.section][picker.target.item]
					else
						settings = getUnitSettings(unit)[picker.target.item]
					end
					local value = slider:GetPosition()
								
					picker.target:SetText(tostring(value))
					
					settings[picker.target.setting] = value
					layoutFrames(unit, false)
				end
				
			end
			
			
			
			-- anchor picker changed
			local function anchorChanged(abox, anchor)
				local unit = _configFrame.activeUnit
				if (unit) then
					local settings
					if (abox.target.section) then
						settings = getUnitSettings(unit)[abox.target.section][abox.target.item]
					else
						settings = getUnitSettings(unit)[abox.target.item]
					end
					
					abox.target:SetText(anchor)
					settings[abox.target.setting] = anchor
					
					layoutFrames(unit, false)			
				end
				abox:SetVisible(false)
			end
			
			_configFrame.IndicatorPicker.Event.SelectedIndexChanged = indChanged
			_configFrame.TextValuePicker.Event.SelectedIndexChanged = textValueChanged			
			_configFrame.BgColorPicker.Event.ColorChanged = bgcolorChanged
			_configFrame.NumberPicker.slider.Event.SliderChange = numberChanged
			_configFrame.AnchorPicker.Event.AnchorChanged = anchorChanged
			
			-- reset unit bar settings
			local function resetClick(reset)
				local unit = _configFrame.activeUnit
				if (unit) and (unit ~= "defaults") then
					local settings
					if (reset.section) then
						settings = _aprofile.frames[unit][reset.section][reset.item]
					else
						settings = _aprofile.frames[unit][reset.item]
					end
					
					settings.show = nil
					settings.margin = nil
					settings.bgcolor =nil
					settings.indicator = nil
					settings.cyratio = nil
					settings.fgcolor = nil
					settings.margin = nil
					settings.xoff = nil
					settings.yoff = nil
					settings.text = nil
					settings.fontsize = nil
					settings.anchor = nil
					settings.tooltip = nil
					settings.cxy = nil
					settings.align = nil
					
					layoutFrames(unit, false)
					
					 _loadUnit(unit)
				end
			end
			local function resetActionClick(reset)
				if (not Inspect.System.Secure()) then
					local unit = _configFrame.activeUnit
					if (unit and (unit ~= "defaults")) then
						_aprofile.frames[unit].actions[reset.item] = nil
						_configFrame.FrameBarControls.actions[reset.item].macro:SetText(L["default action"])
						
						updateMacros(unit)
					end
				end
			end
			local function applyActionClick(apply)
				if (not Inspect.System.Secure()) then
					local unit = _configFrame.activeUnit
					if (unit) then
						local m = trim((_configFrame.FrameBarControls.actions[apply.item].macro:GetText()))
						
						if (unit == "defaults") then
							_aprofile.actions[apply.item] = m
						else
							_aprofile.frames[unit].actions[apply.item] = m
						end
						
						
						updateMacros(unit)
					
					end
				end
			end
			
			
			if (true) then 	--  grids
				local vpbars = myUI:CreateFrame("myUI_Viewport", tpage)
				vpbars:SetPoint("TOPLEFT", tpage, "TOPLEFT", margin, margin)
				vpbars:SetPoint("BOTTOMRIGHT", tpage, "BOTTOMRIGHT", -margin/2, -margin)
				
				local content = vpbars.Content
				
				content:SetHeight(1100)
				--content:SetBackgroundColor(1,0,0,0.1)  -- just to see how big it is when testing
				
				local barControls = {} --	unitframe={}, portrait = {}, titlebar = {}, healthbar = {}, powerbar = {}, chargebar = {}, etc.
					
				if (true) then 	-- bars				
					local colheaders = {
						{L["show"] , 120}, 
						{L["margin"] , 50}, 
						{L["bgcolor"] , 60}, 
						{L["indicator"] , 40}, 
						{L["indicator2"], 110},
						{L["height"] , 140},
					}
					
					local anchor = content
					local anchorTo = "TOPLEFT"
					local offy = 10
					
					local ghdr = myUI:CreateFrame("Text", content)
					ghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					ghdr:SetText(L["bars"])
					ghdr:SetWidth(80)
					ghdr:SetLayer(4)
					
					local cy = ghdr:GetHeight()
					
					ghdr = myUI:CreateFrame("Frame", content)
					ghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					ghdr:SetPoint("RIGHTCENTER", content, "TOPRIGHT", 0, offy)
					ghdr:SetBackgroundColor(rgba(myuicolors.Hover))
					ghdr:SetHeight(cy)
					ghdr:SetLayer(0)
					
					for _, hdr in ipairs(colheaders) do
						local h = myUI:CreateFrame("Text", content)
						h:SetText(hdr[1])
						h:SetLayer(4)
						h:SetPoint("LEFTCENTER", anchor, anchorTo, hdr[2], offy)
						
						offy = 0
						anchor = h
						anchorTo = "LEFTCENTER"
					end
					
					local rowheaders = {"unitframe", "portrait","titlebar", "healthbar", "powerbar", "chargebar"}
					
					yoff = anchor:GetBottom() - content:GetTop() + 4
					anchor = content
					anchorTo = "TOPLEFT"
					
					
					for i, row in ipairs(rowheaders) do
						barControls[row] = {}
					
						local h = myUI:CreateFrame("Text", content)
						h:SetText(L[row])
						h:SetPoint("TOPLEFT", anchor, anchorTo, 0, yoff)
						h.item = row
						barControls[row].header = h
						
						local show = myUI:CreateFrame("RiftCheckbox", content)
						show:SetPoint("LEFTCENTER", h, "LEFTCENTER", colheaders[1][2]+4, 0)					
						show:SetVisible((i ~= 1) and (i ~= 4))
						show.item = row
						show.Event.CheckboxChange = showChecked
						
						if (i == 1) or (1 == 4) then
							show:SetVisible(false)
						else
							barControls[row].show = show
						end
						
						local marg = myUI:CreateFrame("Text", content)
						marg:SetWidth(40)
						marg:SetText(" ")
						marg:SetBackgroundColor(rgba(myuicolors.Selected))
						marg:SetPoint("LEFTCENTER", show, "LEFTCENTER", colheaders[2][2], 0)
						marg.min = 0
						marg.max = 10
						marg.item = row
						marg.setting = "margin"
						marg.Event.LeftDown = numberClick
						
						barControls[row].margin = marg
					
						local clr = myUI:CreateFrame("Frame", content)
						clr:SetSize(14)
						clr:SetPoint("LEFTCENTER", marg, "LEFTCENTER", colheaders[3][2]+4, 0)
						clr.Event.LeftDown = colorClick					
						clr:SetBackgroundColor(0,0,0,1)
						clr.item = row
						
						barControls[row].bgcolor = clr
						
						local ind = myUI:CreateFrame("Text", content)
						ind:SetWidth(100)
						ind:SetText(" ") -- establish height
						ind:SetPoint("LEFTCENTER", clr, "LEFTCENTER", colheaders[4][2], 0)
						ind:SetBackgroundColor(rgba(myuicolors.Selected))
						ind.item = row
						
						ind.Event.LeftDown = indClick
						
						barControls[row].indicator = ind
						
						local ind2 = myUI:CreateFrame("Text", content)
						ind2:SetWidth(100)
						ind2:SetText(" ")
						ind2:SetPoint("LEFTCENTER", ind, "LEFTCENTER", colheaders[5][2], 0)
						ind2:SetBackgroundColor(rgba(myuicolors.Selected))
						ind2.item = row
						ind2.setting = "indicator2"
						
						ind2.Event.LeftDown = indClick
						
						barControls[row].indicator2 = ind2
						
						local hgt = myUI:CreateFrame("RiftSlider", content)
						hgt:SetWidth(80)
						hgt:SetPoint("LEFTCENTER", ind2, "LEFTCENTER", colheaders[6][2]-4, 8)
						hgt:SetRange(1, 10)
						hgt.item = row
						
						
						if (i > 2) then
							barControls[row].cyratio = hgt
							hgt.Event.SliderChange = cyratioChanged	
						elseif (i == 2) then
							hgt:SetRange(1,2)
							barControls[row].align = hgt
							hgt.Event.SliderChange = alignChanged
						else
							hgt:SetVisible(false)
						end
						
						local reset = myUI:CreateFrame("myUI_Button", content)
						reset:SetText(L["reset"])
						reset:SetPoint("LEFTCENTER", hgt, "RIGHTCENTER", 16, -8)
						reset.item = row
						reset.Event.LeftPress = resetClick
						
						barControls[row].reset = reset
						
						anchor = h
						anchorTo = "BOTTOMLEFT"
						yoff = 4
					end
									
				end
				if (true) then	-- bar text
					local colheaders = {
						{L["fgcolor"], 114},
						{L["fontsize"], 48},
						{L["xoff"], 60},
						{L["yoff"], 68},
						{L["indicator"], 68},
						{L["text"], 118},
					}
					
					local anchor = content
					local offy = 48 + barControls.chargebar.header:GetTop() - content:GetTop()	-- vert spacing between bar properties and text properties
					
					local ghdr = myUI:CreateFrame("Text", content)
					ghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					ghdr:SetText(L["text"])
					ghdr:SetWidth(80)
					ghdr:SetLayer(4)
					
					local cy = ghdr:GetHeight()
					anchor = ghdr
					
					local bghdr = myUI:CreateFrame("Frame", content)
					bghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					bghdr:SetPoint("RIGHTCENTER", content, "TOPRIGHT", 0, offy)
					bghdr:SetBackgroundColor(rgba(myuicolors.Hover))
					bghdr:SetHeight(cy)
					bghdr:SetLayer(0)
					
					offy = 0
					
					for _, hdr in ipairs(colheaders) do
						local h = myUI:CreateFrame("Text", content)
						h:SetText(hdr[1])
						h:SetPoint("LEFTCENTER", anchor, "LEFTCENTER", hdr[2], offy)
						
						offy = 0
						anchor = h
					end
					
					
					
					local rowheaders = {"titlebar", "healthbar", "powerbar", "chargebar", "extra"}
					local sub_rows = {"_text1", "_text2"}
					
					offy = 4
					local anchorTo = "BOTTOMLEFT"
					anchor = ghdr
					
					for i, prerow in ipairs(rowheaders) do
						for j, subrow in ipairs(sub_rows) do
							local row = prerow .. subrow
							barControls[row] = {}
				
							local h = myUI:CreateFrame("Text", content)
							h:SetText(L[prerow] .. " " .. L[subrow])
							h:SetPoint("TOPLEFT", anchor, anchorTo, 0, offy)
							h.item = row
							
							barControls[row].header = h
							
							local clr = myUI:CreateFrame("Frame", content)
							clr:SetSize(14)
							clr:SetPoint("LEFTCENTER", h, "LEFTCENTER", colheaders[1][2] + 4, 0)
							clr.Event.LeftDown = colorClick					
							clr:SetBackgroundColor(0,0,0,1)
							clr.item = row
							
							barControls[row].fgcolor = clr
							
							local fsize = myUI:CreateFrame("Text", content)
							fsize:SetWidth(40)
							fsize:SetText(" ")
							fsize:SetBackgroundColor(rgba(myuicolors.Selected))
							fsize:SetPoint("LEFTCENTER", clr, "LEFTCENTER", colheaders[2][2],  0)
							fsize.min = 4
							fsize.max = 60
							fsize.setting = "fontsize"
							fsize.item = row						
							fsize.Event.LeftDown = numberClick
							
							barControls[row].fontsize = fsize
							
							local xoff = myUI:CreateFrame("Text", content)
							xoff:SetWidth(50)
							xoff:SetText(" ")
							xoff:SetBackgroundColor(rgba(myuicolors.Selected))
							xoff:SetPoint("LEFTCENTER", fsize, "LEFTCENTER", colheaders[3][2], 0)
							xoff.min =-200
							xoff.max = 200
							xoff.item = row
							xoff.setting = "xoff"
							xoff.Event.LeftDown = numberClick
							
							barControls[row].xoff = xoff
							
							local yoff = myUI:CreateFrame("Text", content)														
							yoff:SetWidth(50)
							yoff:SetText(" ")
							yoff:SetBackgroundColor(rgba(myuicolors.Selected))
							yoff:SetPoint("LEFTCENTER", xoff, "LEFTCENTER", colheaders[4][2], 0)
							yoff.min = -200
							yoff.max = 200
							yoff.item = row
							yoff.setting = "yoff"
							yoff.Event.LeftDown = numberClick
							
							barControls[row].yoff = yoff
							
							local ind = myUI:CreateFrame("Text", content)
							ind:SetWidth(100)
							ind:SetText(" ") -- establish height
							ind:SetPoint("LEFTCENTER", yoff, "LEFTCENTER", colheaders[5][2], 0)
							ind:SetBackgroundColor(rgba(myuicolors.Selected))
							ind.item = row							
							ind.Event.LeftDown = indClick
							
							barControls[row].indicator = ind
							
							local txt = myUI:CreateFrame("Text", content)
							txt:SetWidth(100)
							txt:SetText(" ")
							txt:SetPoint("LEFTCENTER", ind, "LEFTCENTER", colheaders[6][2], 0)
							txt:SetBackgroundColor(rgba(myuicolors.Selected))
							txt.item = row
							txt.Event.LeftDown = textClick
							
							barControls[row].text = txt
							
							local reset = myUI:CreateFrame("myUI_Button", content)
							reset:SetText(L["reset"])
							reset:SetPoint("LEFTCENTER", txt, "RIGHTCENTER", 16, 0)
							reset.item = row
							reset.Event.LeftPress = resetClick
							
							barControls[row].reset = reset
							
							anchorTo = "BOTTOMLEFT"
							anchor =h				
						end						
					end
				end
				
				if (true) then 	-- icons
					local colheaders = {
						{L["show"], 120},
						{L["cxy"], 50},
						{L["xoff"], 60},
						{L["yoff"], 60},
						{L["anchor"], 70},
						{L["tooltip"], 138}						
					}
					
					local anchor = content
					local offy = 48 + barControls.extra_text2.header:GetTop() - content:GetTop()	-- vert spacing between bar properties and text properties
					
					local ghdr = myUI:CreateFrame("Text", content)
					ghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					ghdr:SetText(L["icons"])
					ghdr:SetWidth(80)
					ghdr:SetLayer(4)
					
					local cy = ghdr:GetHeight()
					anchor = ghdr
					
					local bghdr = myUI:CreateFrame("Frame", content)
					bghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					bghdr:SetPoint("RIGHTCENTER", content, "TOPRIGHT", 0, offy)
					bghdr:SetBackgroundColor(rgba(myuicolors.Hover))
					bghdr:SetHeight(cy)
					bghdr:SetLayer(0)
					
					offy = 0
					
					for _, hdr in ipairs(colheaders) do
						local h = myUI:CreateFrame("Text", content)
						h:SetText(hdr[1])
						h:SetPoint("LEFTCENTER", anchor, "LEFTCENTER", hdr[2], offy)
						
						offy = 0
						anchor = h
					end
				
					local rowheaders = {"dead", "faction", "loot", "player", "pvp","role", "tagged", "tier","vitality", "ready"}
				
					offy = 4
					local anchorTo = "BOTTOMLEFT"
					anchor = ghdr
					
					local icons = {}
					barControls.icons = icons
					
					for i, row in ipairs(rowheaders) do
						icons[row] = {}
						
						local h = myUI:CreateFrame("Text", content)
						h:SetText(L[row])
						h:SetPoint("TOPLEFT", anchor, anchorTo, 0, yoff)
						h.item = row
						icons[row].header = h
						
						local show = myUI:CreateFrame("RiftCheckbox", content)
						show:SetPoint("LEFTCENTER", h, "LEFTCENTER", colheaders[1][2]+4, 0)					
						show.item = row						
						show.Event.CheckboxChange = showIconChecked
						icons[row].show = show
						
						local cxy = myUI:CreateFrame("Text", content)
						cxy:SetWidth(40)
						cxy:SetText(" ")
						cxy:SetBackgroundColor(rgba(myuicolors.Selected))
						cxy:SetPoint("LEFTCENTER", show, "LEFTCENTER", colheaders[2][2],  0)
						cxy.min = 8
						cxy.max = 128
						cxy.setting = "cxy"
						cxy.section = "icons"
						cxy.item = row						
						cxy.Event.LeftDown = numberClick
						icons[row].cxy = cxy
						
						local xoff = myUI:CreateFrame("Text", content)
						xoff:SetWidth(50)
						xoff:SetText(" ")
						xoff:SetBackgroundColor(rgba(myuicolors.Selected))
						xoff:SetPoint("LEFTCENTER", cxy, "LEFTCENTER", colheaders[3][2], 0)
						xoff.min =-200
						xoff.max = 200
						xoff.item = row
						xoff.setting = "xoff"
						xoff.section = "icons"
						xoff.Event.LeftDown = numberClick
						
						icons[row].xoff = xoff
						
						local yoff = myUI:CreateFrame("Text", content)														
						yoff:SetWidth(50)
						yoff:SetText(" ")
						yoff:SetBackgroundColor(rgba(myuicolors.Selected))
						yoff:SetPoint("LEFTCENTER", xoff, "LEFTCENTER", colheaders[4][2], 0)
						yoff.min = -400
						yoff.max = 400
						yoff.item = row
						yoff.setting = "yoff"
						yoff.section = "icons"
						yoff.Event.LeftDown = numberClick
						
						icons[row].yoff = yoff
						
						local anch = myUI:CreateFrame("Text", content)
						anch:SetWidth(120)
						anch:SetText(" ")
						anch:SetBackgroundColor(rgba(myuicolors.Selected))
						anch:SetPoint("LEFTCENTER", yoff, "LEFTCENTER", colheaders[5][2], 0)
						anch.item = row
						anch.setting = "anchor"
						anch.section = "icons"
						anch.Event.LeftDown = anchorClick
						
						icons[row].anchor = anch
					
						local ind = myUI:CreateFrame("Text", content)
						ind:SetWidth(100)
						ind:SetText(" ") -- establish height
						ind:SetPoint("LEFTCENTER", anch, "LEFTCENTER", colheaders[6][2], 0)
						ind:SetBackgroundColor(rgba(myuicolors.Selected))
						ind.item = row			
						ind.section = "icons"
						ind.setting = "tooltip"
						ind.Event.LeftDown = textClick
						
						icons[row].tooltip = ind
						
						local reset = myUI:CreateFrame("myUI_Button", content)
						reset:SetText(L["reset"])
						reset:SetPoint("LEFTCENTER", ind, "RIGHTCENTER", 16, 0)
						reset.item = row
						reset.section = "icons"
						reset.Event.LeftPress = resetClick
						
						icons[row].reset = reset
						
						anchorTo = "BOTTOMLEFT"
						anchor =h	
					end
				end
				
				if (true) then 	-- actions
					local colheaders = {{L["macro"], 120}}
					local anchor = content
					local offy = 48 + barControls.icons.ready.header:GetTop() - content:GetTop()
					
					local ghdr = myUI:CreateFrame("Text", content)
					ghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					ghdr:SetText(L["actions"])
					ghdr:SetWidth(100)
					ghdr:SetLayer(4)
					
					local cy = ghdr:GetHeight()
					anchor = ghdr
					
					local bghdr = myUI:CreateFrame("Frame", content)
					bghdr:SetPoint("LEFTCENTER", content, "TOPLEFT", 0, offy)
					bghdr:SetPoint("RIGHTCENTER", content, "TOPRIGHT", 0, offy)
					bghdr:SetBackgroundColor(rgba(myuicolors.Hover))
					bghdr:SetHeight(cy)
					bghdr:SetLayer(0)
					
					offy = 0
					
					for _, hdr in ipairs(colheaders) do
						local h = myUI:CreateFrame("Text", content)
						h:SetText(hdr[1])
						h:SetPoint("LEFTCENTER", anchor, "LEFTCENTER", hdr[2], offy)
						offy = 0
						anchor = h
					end
					
					local rowheaders = {"LeftDown", "RightDown", "MiddleDown", "Mouse4Down", "Mouse5Down"}
					offy = 4
					local anchorTo = "BOTTOMLEFT"
					anchor = ghdr
					
					local actions = {}
					barControls.actions = actions
					
					for i, row in ipairs(rowheaders) do
						actions[row] = {}
						
						local h = myUI:CreateFrame("Text", content)
						h:SetText(L[row])
						h:SetPoint("TOPLEFT", anchor, anchorTo, 0, yoff)
						h.item = row
						actions[row].header = h
					
						local m = myUI:CreateFrame("RiftTextfield", content)
						m:SetPoint("LEFTCENTER", h, "LEFTCENTER",  colheaders[1][2], 0)
						m:SetHeight(h:GetHeight())
						m:SetBackgroundColor(rgba(myuicolors.Selected))
						m:SetWidth(400)
						
						actions[row].macro = m
					
						local apply = myUI:CreateFrame("myUI_Button", content)
						apply:SetText(L["apply"])
						apply:SetPoint("LEFTCENTER", m, "RIGHTCENTER", 16, 0)
						apply.item = row
						apply.section = "actions"
						apply.Event.LeftPress = applyActionClick
						
						actions[row].apply = apply
					
					
						local reset = myUI:CreateFrame("myUI_Button", content)
						reset:SetText(L["reset"])
						reset:SetPoint("LEFTCENTER", apply, "RIGHTCENTER", 16, 0)
						reset.item = row
						reset.section = "actions"
						reset.Event.LeftPress = resetActionClick
						
						actions[row].reset = reset
						
					
						anchorTo = "BOTTOMLEFT"
						anchor =h	
					end
				
				
					local offy = 20 + barControls.actions.Mouse5Down.header:GetBottom() - content:GetTop()
					
					local instr = myUI:CreateFrame("Text", content)
					instr:SetWordwrap(true)					
					instr:SetPoint("TOPLEFT", content, "TOPLEFT", 0, offy)
					instr:SetPoint("TOPRIGHT", content, "TOPRIGHT", 0, offy)
					instr:SetText(L["macro instructions"])
					
					
				end
				
				_configFrame.FrameBarControls = barControls
			end
				
			
			
			
		end
		
		

		
		tstrip.Event.TabSelected = function(tstrip, value)
			for _, tp in pairs(_configFrame.TabPages) do
				tp:SetVisible(false)
			end
			
			if (type(value) == "string") then
				if (value == L["Global"]) then	
					
					_configFrame.TabPages["global"]:SetVisible(true)				
				end
			elseif (type(value) == "table") then				
				if (value.unit) then					
					_loadUnit(value.unit)					
					_configFrame.TabPages["unit"]:SetVisible(true)
					
					
				end
			end
		end
		
		_configFrame.LoadProfile = function(self)
			self._loading = true
			
			local combat = Inspect.System.Secure()
			
			for _, box in ipairs(self.ColorBoxes) do
				self.ColorPicker:SetVisible(false)
				
				box:SetBackgroundColor(rgba(_aprofile.colors[box.indicator][box.value]))
			end
			
			for k, v in pairs(_aprofile.frames) do
				_configFrame.UnitChecks[k]:SetChecked(ifnil(v.enabled, false))
				--_configFrame.UnitChecks[k]:SetEnabled(not combat)
				
			end
			
			self._loading = false
		end
		
		_configFrame.SetCombatControls = function(self, incombat)
			if (self.activeUnit) then
				local mouseover = string.find(self.activeUnit, "mouseover")
				local defaults = (self.activeUnit == "defaults")
				
				local ok = ((not mouseover) and (not incombat))
				
				for action, ctrls in pairs(self.FrameBarControls.actions) do
					ctrls.apply:SetEnabled(ok)
					ctrls.reset:SetEnabled(ok and (not defaults))
				end
			end
		end
		
		_configFrame.TabPages = tpages
		_configFrame.TabStrip = tstrip
		
		_configFrame:LoadProfile()
	end
	return _configFrame
end
function unitframes:ShowMovers()
	for _, uf in pairs(_frames) do
		if (uf.enabled) then
			uf.mover:SetVisible(true)
			uf.mover.dragging = false
			uf.mover.sizer.sizing = false
		end
	end
end
function unitframes:HideMovers()
	for _, uf in pairs(_frames) do
		uf.mover:SetVisible(false)
		uf.mover.dragging = false
		uf.mover.sizer.sizing = false
	end
end

--

myUI.UnitFrames = unitframes -- we'll expose this one to other addons via myUI core.