local myUI = ...
local L = myUI.data.language

local pprint = myUI.Utility.pprint
local iif = myUI.Utility.iif
local rgba = myUI.Utility.rgba
local ifnil = myUI.Utility.ifnil
local tohex = myUI.Utility.tohex

myUI.Colors = {
	Mover = {0.6,0,0,0.2},
	Hover = {30/255,144/255,1,0.4},	-- DodgerBlue	
	Selected = {0,0,0,0.4},
	Border = {0.5,0.5,0.5,1},
	Transparent = {0,0,0,0},
	TextArea = {0,0,0,1},
	ActiveTab = {0,0,0,0.4},
	DisabledText = {0.5, 0.5, 0.5, 1},
	NormalText = {1,1,1,1},
	Button = {0,0,0,0.8},
	ButtonDown = {20/255,122/255,0.8,1},
	TooltipBg = {0,0,0,0.8},
	TooltipTitle = {30/255,144/255,1,1},
	TooltipText = {1,1,1,1},
}

local clrMover = myUI.Colors.Mover
local clrHover = myUI.Colors.Hover
local clrSelected = myUI.Colors.Selected
local clrBorder = myUI.Colors.Border
local clrTransparent = myUI.Colors.Transparent
local clrTextArea = myUI.Colors.TextArea
local clrActiveTab = myUI.Colors.ActiveTab
local clrDisabled = myUI.Colors.DisabledText
local clrNormal = myUI.Colors.NormalText
local clrButton = myUI.Colors.Button
local clrButtonDown = myUI.Colors.ButtonDown

myUI.Layers = {
	Configuration = 100,
	Tooltip = 1000,
	UnitFrame = 10,
	
}

-- tooltip
local _ttip
local _ttipid = 0

function myUI:ShowTooltip(caller, title, message, cx)
	if (not caller) then return end
	
	if (not _ttip) then
		_ttip = myUI:CreateFrame("myUI_BorderBox")
		_ttip:SetLayer(myUI.Layers.Tooltip)
		_ttip:SetBackgroundColor(rgba(myUI.Colors.TooltipBg))
		
		_ttip.title = myUI:CreateFrame("Text", _ttip)
		_ttip.title:SetLayer(2)
		_ttip.title:SetFontColor(rgba(myUI.Colors.TooltipTitle))
		_ttip.title:SetFontSize(_ttip.title:GetFontSize()+2)
		_ttip.title.customFontSize = true
		
		_ttip.msg = myUI:CreateFrame("Text", _ttip)
		_ttip.msg:SetLayer(2)
		_ttip.msg:SetFontColor(rgba(myUI.Colors.TooltipText))
		_ttip.msg:SetWordwrap(true)
		
		
		_ttip.msg:SetPoint("BOTTOMLEFT", _ttip, "BOTTOMLEFT", 2, -2)
		_ttip.msg:SetPoint("BOTTOMRIGHT", _ttip, "BOTTOMRIGHT", -2, -2)
		
		_ttip.title:SetPoint("BOTTOMLEFT", _ttip.msg, "TOPLEFT", 0, -2)
		
		
		
	end

	if (title) then 
		_ttip.title:ClearHeight()
		_ttip.title:SetText(title)
	else
		_ttip.title:SetHeight(0)
	end
	if (message) then
		_ttip.msg:ClearHeight()
		_ttip.msg:SetText(message)
	else
		_ttip.msg:SetHeight(0)
	end
	
	cx = ifnil(cx, 240)
	
	
	_ttip:SetWidth(cx)
	_ttip:SetHeight(_ttip.msg:GetHeight() + 2 + _ttip.title:GetHeight() +4)
	
	local m = Inspect.Mouse()
	m.x = m.x + 8
	m.y = m.y - 8
	
	local maxx = UIParent:GetWidth()
	local maxy = UIParent:GetHeight()
	
	local cy = _ttip:GetHeight()
	if ((m.x + cx) > maxx) then
		m.x = maxx - cx - 16
	end
	if ((m.y + cy) > maxy) then
		m.y = maxy - cx - 16
	elseif (m.y - _ttip:GetHeight() < 0) then
		m.y = 8 + _ttip:GetHeight()
	end
	
	_ttipid = _ttipid + 1
	caller._ttipid = _ttipid
	
	_ttip:SetPoint("BOTTOMLEFT", UIParent, "TOPLEFT", m.x, m.y)
	
	
	_ttip:SetVisible(true)	
end

function myUI:HideTooltip(caller)
	if (not caller) or (not caller._ttipid) then return end
	
	if (caller._ttipid >= _ttipid) then
		caller._ttipid = nil
		_ttip:SetVisible(false)
	end
end
-- BorderBox widget --

local function _setBorderColorBBox(self, r, g, b, a)
	for _, border in pairs(self._borders) do
		border:SetBackgroundColor(r, g, b, a)
	end
end
local function _setBorderSizeBBox(self, size)
	if (size) and (size > 0) then
		self._borders.left:SetPoint("TOPLEFT", self, "TOPLEFT")
		self._borders.left:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT")
		self._borders.left:SetWidth(size)
		
		self._borders.top:SetPoint("TOPLEFT", self, "TOPLEFT")
		self._borders.top:SetPoint("TOPRIGHT", self, "TOPRIGHT")
		self._borders.top:SetHeight(size)
		
		self._borders.right:SetPoint("TOPRIGHT", self, "TOPRIGHT")
		self._borders.right:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT")
		self._borders.right:SetWidth(size)
		
		self._borders.bottom:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT")
		self._borders.bottom:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT")
		self._borders.bottom:SetHeight(size)
	
		self._borderSize = size
	end
end
local function _getBorderSizeBBox(self)
	return self._borderSize
end
local function _bordersBBox(bbox)
	return pairs(bbox._borders)
end




local function _createBorderBox(parent)
	local bg = myUI:CreateFrame("Mask", parent)

	bg._borders = {
		left = myUI:CreateFrame("Frame", bg),
		top = myUI:CreateFrame("Frame", bg),
		right = myUI:CreateFrame("Frame", bg),
		bottom = myUI:CreateFrame("Frame", bg)
	}
	
	bg._borders.left:SetLayer(1)
	bg._borders.top:SetLayer(1)
	bg._borders.right:SetLayer(1)
	bg._borders.bottom:SetLayer(1)
	
	bg.SetBorderColor = _setBorderColorBBox
	bg.SetBorderSize = _setBorderSizeBBox
	bg.GetBorderSize = _getBorderSizeBBox	
	bg.Borders = _bordersBBox
	
	bg:SetBorderColor(rgba(clrBorder))
	bg:SetBorderSize(2)
	
	return bg	
end



-- Listbox widget --

local _recycledListBoxItems = {}

local function _itemsLB(lbox)
	return ipairs(lbox._items)
end

local function _onMouseInItemLB(item)
	item:SetBackgroundColor(rgba(item:GetParent():GetParent():GetParent()._clrHover))
end

local function _onMouseOutItemLB(item)	
	if (item._selected) then
		item:SetBackgroundColor(rgba(item:GetParent():GetParent():GetParent()._clrSelected))
	else
		item:SetBackgroundColor(rgba(clrTransparent))
	end
end

local function _setItemSelectedLB(item, selected, suspendNotification)
	if (selected ~= item:GetSelected()) then
		local content = item:GetParent()
		local lbox = content:GetParent():GetParent()
		
		if (selected) then
			item:SetBackgroundColor(rgba(lbox._clrSelected))
			item._selected = true
		else
			item:SetBackgroundColor(rgba(clrTransparent))
			item._selected = false
		end
		
		if (content ~= UIParent) then
			
			
			if (selected) then
				if (not lbox:GetMultiSelect()) then
					-- can only have one item selected so
					for _, item2 in lbox:Items() do
						if (item2 ~= item) then
							if (item2:GetSelected()) then
								item2:SetSelected(false, true)
							end
						end
					end
				end
			end
			if (not suspendNotification) and (lbox.Event.SelectedIndexChanged) then
				lbox.Event.SelectedIndexChanged(lbox, lbox:GetSelectedIndexes())
			end
		end
		
	end
end

local function _getSelectedIndexesLB(lbox)
	local indexes = {}
	
	for i, item in ipairs(lbox._items) do
		if (item:GetSelected()) then
			table.insert(indexes, i)
		end
	end
	
	return indexes
end

local function _selectItemLB(lbox, index, selected)
	selected = ifnil(selected, true)
	
	if (index) and (index > 0) and (index <= #lbox._items) then
		lbox._items[index]:SetSelected(selected)
	end
end

local function _getItemLB(lbox, index)
	if (index > 0) and (index <= #lbox._items) then
		return lbox._items[index]._value
	end
end

local function _setItemLB(lbox, index, data)
	if (index > 0) and (index <= #lbox.items) then
		if (data.tostring) then
			lbox._items[index]._text:SetText(data:tostring())
		else
			lbox._items[index]._text:SetText(tostring(data))
		end
		lbox._items[index]._value = data
		lbox._items[index]:SetHeight(lbox._items[index]._text:GetHeight() + 8)
	end
end
local function _getItemCountLB(lbox)
	return #lbox._items
end

local function _getItemSelectedLB(item)
	return item._selected
end

local function _clickItemLB(item)
	item:SetSelected(not item:GetSelected())
end



local function _getListBoxItem(parent)
	local item
	if (#_recycledListBoxItems > 0) then
		item = table.Remove(_recycledListBoxItems)
		item:SetParent(parent)
	else
		item = myUI:CreateFrame("Frame", parent)
		item._text = myUI:CreateFrame("Text", item)
		item._text:SetPoint("LEFTCENTER", item, "LEFTCENTER", 4, 0)
		item._text:SetPoint("RIGHTCENTER", item, "RIGHTCENTER", -4, 0)
		
		item.Event.MouseIn = _onMouseInItemLB
		item.Event.MouseOut = _onMouseOutItemLB
		item.Event.LeftDown = _clickItemLB
		
		item._selected = false
		item.SetSelected = _setItemSelectedLB
		item.GetSelected = _getItemSelectedLB
	end
	
	item:SetBackgroundColor(rgba(clrTransparent))
	item._text:SetWordwrap(true)
	return item
end
local function _recycleListBoxItem(item)
	item:SetVisible(false)
	item._text:SetText("")
	item:ClearAll()	
	item._value = nil	
	item:SetSelected(false, true)	
end

local function _layoutItemsLB(lbox)
	if (#lbox._items == 0) then
		lbox.Content:SetHeight(4)
		lbox.Content:SetPoint("TOPCENTER", lbox.Viewport, "TOPCENTER", 0, lbox.Viewport:GetBorderSize())
		lbox.Scrollbar:SetEnabled(false)
	else
		local cy = lbox.Viewport:GetBorderSize()
		
		for _, item in ipairs(lbox._items) do
			item:SetHeight(item._text:GetHeight() + lbox._spacing)
			cy = cy + item:GetHeight()
		end
		
		lbox.Content:SetHeight(cy)
		local offy = cy - lbox.Viewport:GetHeight() 
		if (offy > 0) then			
			lbox.Scrollbar:SetRange(0, offy)
			lbox.Scrollbar:SetEnabled(true)			
			
		else
			lbox.Content:SetPoint("TOPCENTER", lbox.Viewport, "TOPCENTER", 0, lbox.Viewport:GetBorderSize())
			lbox.Scrollbar:SetEnabled(false)
		end
	end
end

local function _setSpacingLB(lbox, spacing)
	if (spacing) and (spacing >= 0) then
		lbox._spacing = spacing
		
		_layoutItemsLB(lbox)
	end
end

local function _onScrollLB(sb)
	local lbox = sb:GetParent()
	local content = lbox.Content
	local vp = lbox.Viewport
	
	local offy = vp:GetBorderSize() - sb:GetPosition()
	content:SetPoint("TOPCENTER", vp, "TOPCENTER", 0, offy)
	
	
end

local function _onViewportSizeLB(viewport)
	local lbox = viewport:GetParent()
	local content = lbox.Content
	
	content:SetWidth(viewport:GetWidth() - (2 * viewport:GetBorderSize()))
	_layoutItemsLB(lbox)
end



local function _addItemLB(lbox, data)
	if (data) then
		local content = lbox.Content
		local item = _getListBoxItem(content)
		
		if (data.tostring) then
			item._text:SetText(data:tostring())
		else
			item._text:SetText(tostring(data))
		end
		item:SetHeight(item._text:GetHeight() + 8)
		
		item._value = data
		
		table.insert(lbox._items, item)
		
		local n = #lbox._items
		
		local anchor = iif(n > 1, lbox._items[n-1], content)
		local anchorTo1 = iif(n > 1, "BOTTOMLEFT", "TOPLEFT")
		local anchorTo2 = iif(n > 1, "BOTTOMRIGHT", "TOPRIGHT")
		
		item:SetPoint("TOPLEFT", anchor, anchorTo1)
		item:SetPoint("TOPRIGHT", anchor, anchorTo2)
	
		_layoutItemsLB(lbox)
	end
end

local function _insertItemLB(lbox, index, data)
	if (index > 0) and (index <= #lbox._items) then
		local content = lbox.Content
		local item = _getListBoxItem(content)
		
		if (data.tostring) then
			item._text:SetText(data:tostring())
		else
			item._text:SetText(tostring(data))
		end
		item:SetHeight(item._text:GetHeight() + 8)
		item._value = data
		
		table.insert(lbox._items, index, item)
		
		if (index == 1) then
			lbox._items[1]:SetPoint("TOPLEFT", content, "TOPLEFT")
			lbox._items[1]:SetPoint("TOPRIGHT", content, "TOPRIGHT")						
		else
			lbox._items[index]:SetPoint("TOPLEFT", lbox._items[index-1], "BOTTOMLEFT")
			lbox._items[index]:SetPoint("TOPRIGHT", lbox._items[index-1], "BOTTOMRIGHT")
		end
		
		lbox._items[index+1]:SetPoint("TOPLEFT", lbox._items[index], "BOTTOMLEFT")
		lbox._items[index+1]:SetPoint("TOPRIGHT", lbox._items[index], "BOTTOMRIGHT")
		
		_layoutItemsLB(lbox)
	end
end

local function _removeItemAtLB(lbox, index)
	if (index) then
		local n = #lbox._items
		if (index > 0) and (index <= n) then
			local selchange = lbox._items[index]:GetSelected()
			
			_recycleListBoxItem(table.remove(lbox._items, index))
			
			if (index < n) then
				if (index == 1) then
						lbox._items[1]:SetPoint("TOPLEFT", lbox.Content, "TOPLEFT")
						lbox._items[1]:SetPoint("TOPRIGHT", lbox.Content, "TOPRIGHT")
				else
						lbox._items[index]:SetPoint("TOPLEFT", lbox._items[index-1], "BOTTOMLEFT")
						lbox._items[index]:SetPoint("TOPRIGHT", lbox._items[index-1], "BOTTOMRIGHT")
				end
			end
		
			if (selchange) and (lbox.Event.SelectedIndexChanged) then
				lbox.Event.SelectedIndexChanged(lbox, lbox:GetSelectedIndexes())
			end
		
			_layoutItemsLB(lbox)
		end
	end
end



local function _getMultiSelectLB(lbox)
	return lbox._multi
end

local function _setMultiSelectLB(lbox, multi)
	lbox._multi = multi
	if (not multi) then
		-- can only have a single item selected
		local selected = false
		for _, item in ipairs(lbox._items) do
			if (item:GetSelected()) then
				if (selected) then
					item:SetSelected(false, true)
				else
					selected = true
				end
			end
		end
	end
end



local function _clearItemsLB(lbox)
	local selchange = (#_getSelectedIndexesLB(lbox) > 0)
	
	for _, item in ipairs(lbox._items) do
		_recycleListBoxItem(item)
	end
	
	lbox._items = {}
	
	if (selchange) and (lbox.Event.SelectedIndexChanged) then
		lbox.Event.SelectedIndexChanged(lbox, lbox:GetSelectedIndexes())
	end
end




local function _setHoverColorLB(lbox, r, g, b, a)
	lbox._clrHover = {r, g, b, a}
end
local function _setSelColorLB(lbox, r, g, b, a)
	lbox._clrSelected = {r, g, b, a}
	
	for _, item in ipairs(lbox._items) do
		if (item:GetSelected()) then
			item:SetBackgroundColor(r,g,b,a)
		end
	end
end

local function _clearSelectionsLB(lbox, suppressEvent)
	if (#lbox._items > 0) then	
		for _, item in lbox:Items() do
			if (item:GetSelected()) then
				item:SetSelected(false, true)
			end		
		end
	
		if (lbox.Event.SelectedIndexChanged) and (not suppressEvent) then
			lbox.Event.SelectedIndexChanged(lbox, {})
		end
	end
end

local function _createListbox(parent)
	local f = myUI:CreateFrame("Frame", parent)
	f._spacing = 0	
	f._clrHover = clrHover
	f._clrSelected = clrSelected
	
	local sb = myUI:CreateFrame("RiftScrollbar", f)
	sb:SetOrientation("vertical")
	sb:SetPoint("TOPRIGHT", f, "TOPRIGHT", 0, 0)
	sb:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", 0, 0)
	sb:SetEnabled(false)
	sb.Event.ScrollbarChange = _onScrollLB
	
	f.Scrollbar = sb
	
	local bx = myUI:CreateFrame("myUI_BorderBox", f)
	bx:SetPoint("TOPLEFT", f, "TOPLEFT")
	bx:SetPoint("BOTTOMRIGHT", sb, "BOTTOMLEFT", -4, 0)
	
	bx.Event.Size = _onViewportSizeLB
	f.Viewport = bx
	
	local content = myUI:CreateFrame("Frame", bx)
	content:SetPoint("TOPCENTER", bx, "TOPCENTER", 0, bx:GetBorderSize())
	content:SetWidth(bx:GetWidth() - (2 * bx:GetBorderSize())-8)
	content:SetHeight(4)
	
	f.Content = content

	f._multi = false
	f._items = {}
	
	f.SetHoverColor = _setHoverColorLB
	f.SetSelectedColor = _setSelColorLB
	
	f.GetMultiSelect = _getMultiSelectLB
	f.SetMultiSelect = _setMultiSelectLB
	
	f.GetSelectedIndexes = _getSelectedIndexesLB
	f.SetItemSelected = _selectItemLB
	
	f.GetItem = _getItemLB
	f.GetItemCount = _getItemCountLB
	
	f.Items = _itemsLB
	f.AddItem = _addItemLB
	f.RemoveItemAt = _removeItemAtLB
	f.ClearItems = _clearItemsLB
	f.InsertItem = _insertItemLB
	f.SetItemSpacing = _setSpacingLB
	f.ClearSelections = _clearSelectionsLB
	
	f.Event = {
		SelectedIndexChanged = nil -- function(listbox, {selectedIndexes})
	}
	
	return f
end





-- DropdownList widget --
local function _getSelIndexDD(self)
	local indexes = self.List:GetSelectedIndexes()
	if (#indexes > 0) then
		return indexes[1]
	else
		return nil
	end
end
local function _clearItemsDD(self)
	self.Text:SetText("")
	self.List:ClearItems()
end

local function _selectedIndexChangedDD(listbox, indexes)
	local self = listbox:GetParent()
	local selindex
	
	if (indexes) and (#indexes > 0) then		
		selindex = indexes[1]		
		local item = listbox:GetItem(selindex)
		if (item) then
			local sz 
			if (item.tostring) then sz = item:tostring() else sz = tostring(item) end
			self.Text:SetText(sz)
		else
			self.Text:SetText("")
		end
	else
		self.Text:SetText("")
	end
	
	listbox:SetVisible(false)

	if (self.Event.SelectedIndexChanged) then
		self.Event.SelectedIndexChanged(lbox, selindex)
	end
end


local function _leftDownDD(text)
	local list = text:GetParent().List
	
	list:SetVisible(not list:GetVisible())
end

local function _itemsDD(self)
	return self.List:Items()
end
local function _addItemDD(self, data)
	return self.List:AddItem(data)
end
local function _removeAtDD(self, index)
	return self.List:RemoveAt(index)
end

local function _insertItemDD(self, index, data)
	return self.List:InsertItem(index, data)
end
local function _setSelectedDD(self, index, selected)
	return self.List:SetItemSelected(index, selected)
end
local function _getItemDD(self, index)
	return self.List:GetItem(index)
end
local function _getCountDD(self)
	return self.List:GetItemCount()
end

local function _createDropDownList(parent)
	local f = myUI:CreateFrame("Frame", parent)
	local t = myUI:CreateFrame("Text", f)
	local list = myUI:CreateFrame("myUI_Listbox", f)
	list:SetVisible(false)
	
	t:SetPoint("TOPLEFT", f, "TOPLEFT", 2, 2)
	t:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -2, -2)
	
	f:SetBackgroundColor(rgba(clrBorder))
	t:SetBackgroundColor(rgba(clrTextArea))
	
	f:SetSize(200, 24)
	
	list:SetPoint("TOPLEFT", f, "BOTTOMLEFT", 0, 2)
	list:SetPoint("TOPRIGHT", f, "BOTTOMRIGHT", 0, 2)
	list:SetHeight(200)
	
	f.Text = t
	f.List = list
	
	f.Items = _itemsDD
	f.AddItem = _addItemDD
	f.RemoveItemAt =_removeAtDD
	f.ClearItems = _clearItemsDD
	f.InsertItem = _insertItemDD
	
	f.GetSelectedIndex = _getSelIndexDD
	f.SetItemSelected = _setSelectedDD
	
	f.GetItem =_getItemDD
	f.GetItemCount = _getCountDD
	
	f.List.Event.SelectedIndexChanged = _selectedIndexChangedDD
	
	t.Event.LeftDown = _leftDownDD
	
	f.Event = {
		SelectedIndexChanged = nil -- function(combobox,  selIndex)
	}
	
	
	
	return f
end

-- Textbox widget --

local function _getParentTB(tbox)
	return tbox:_getParent():GetParent()
end

local function _onKeyTB(tbox, key)
	local code = string.byte(key)
	if (code == 13) and (tbox.Event.OnEnter) then
		tbox.Event.OnEnter(tbox)
	end
end

local function _setBorderColorTB(tbox, r, g, b, a)
	tbox._bg:SetBackgroundColor(r,g,b,a)
end
local function _getBorderColorTB(tbox)
	return tbox._bg:GetBackgroundColor()
end
local function _setBorderSizeTB(tbox, size)
	if (size) and (size > 0) then
		tbox._bg:SetPoint("TOPLEFT", tbox, "TOPLEFT", -size, -size)
		tbox._bg:SetPoint("BOTTOMRIGHT", tbox, "BOTTOMRIGHT", size, size)
		tbox._bg._borderSize = 2
	end
end
local function _getBorderSizeTB(tbox)
	return tbox._bg._borderSize
end

local function _createTextbox(parent)
	local bg = myUI:CreateFrame("Frame", parent)
	
	local brdr = myUI:CreateFrame("Frame", bg)
	brdr:SetLayer(0)
	brdr:SetBackgroundColor(rgba(clrBorder))
	
	local txt = myUI:CreateFrame("RiftTextfield", bg)
	txt:SetLayer(1)
	txt:SetBackgroundColor(rgba(clrTextArea))
	
	txt._bg = brdr
	
	-- b/c we've doing a 'fake' parent to get layering correct, we need to override GetParent() to get our fake layer's parent
	txt._getParent = txt.GetParent
	
	txt.GetParent = _getParentTB
	txt.SetBorderColor = _setBorderColorTB
	txt.GetBorderColor = _getBorderColorTB
	txt.SetBorderSize = _setBorderSizeTB
	txt.GetBorderSize = _getBorderSizeTB
	
	txt.Event.KeyType = _onKeyTB
	
	txt:SetBorderSize(2)
	txt.Event = {
		OnEnter = nil	
	}
	
	return txt
end





-- Checkbox widget --

local function _getTextCB(chk)
	return chk._label:GetText()
end

local function _setTextCB(chk, text)
	chk._label:SetText(text)
end

local function _getFontCB(chk)
	return chk._label:GetFont()
end

local function _setFontCB(chk, src, file)
	chk._label:SetFont(src, file)
end

local function _getFontSizeCB(chk)
	return chk._label:GetFontSize()
end
local function _setFontSizeCB(chk, size)
	chk._label:SetFontSize(size)
end

local function _getFontColorCB(chk)
	return chk._label:GetFontColor()
end
local function _setFontColorCB(chk, r, g, b, a)
	if (chk:GetEnabled()) then
		chk._label:SetFontColor(r, g, b, a)
		chk._clr = {r,g,b,a}
	end
end
local function _setEnabledCB(chk, enabled)
	if (enabled) then
		chk._label:SetFontColor(rgba(chk._clr))
	else
		chk._label:SetFontColor(rgba(clrDisabled))
	end
	chk:_setEnabled(enabled)
end

local function _createCheckbox(parent)
	local chk = myUI:CreateFrame("RiftCheckbox", parent)
	local txt = myUI:CreateFrame("Text", chk)
	
	txt:SetPoint("LEFTCENTER", chk, "RIGHTCENTER")
	
	chk._label = txt
	chk._clr = clrNormal
	
	chk.SetText = _setTextCB
	chk.GetText = _getTextCB
	
	chk.GetFont = _getFontCB
	chk.SetFont = _setFontCB
	
	chk.GetFontSize = _getFontSizeCB
	chk.SetFontSize = _setFontSizeCB
	
	chk.GetFontColor = _getFontColorCB
	chk.SetFontColor = _setFontColorCB
	
	chk._setEnabled = chk.SetEnabled
	chk.SetEnabled = _setEnabledCB
	
	return chk
end


-- Tabstrip widget --

local function _tabsTS(tstrip) 
	return ipairs(tstrip._tabs)
end

local function _mouseInTS(tab)
	tab:SetBackgroundColor(rgba(clrHover))
end

local function _mouseOutTS(tab)
	if (tab._selected) then
		tab:SetBackgroundColor(rgba(clrActiveTab))
	else
		tab:SetBackgroundColor(rgba(clrTransparent))
	end
end

local function _selectTabTS(tab, selected)
	if (selected ~= tab._selected) then
		tab._selected = selected
		if (selected) then
			tab:SetBackgroundColor(rgba(clrActiveTab))
			
			for _, t in tab:GetParent():Tabs() do
				if (t ~= tab) then t:SetSelected(false) end
			end
			
			local tstrip = tab:GetParent()
			
			tstrip._borders.left:SetPoint("TOPLEFT", tab, "TOPLEFT")
			tstrip._borders.left:SetPoint("BOTTOMLEFT", tab, "BOTTOMLEFT")
			
			tstrip._borders.top:SetPoint("TOPLEFT", tab, "TOPLEFT")
			tstrip._borders.top:SetPoint("TOPRIGHT", tab, "TOPRIGHT")
			
			tstrip._borders.bottom:SetPoint("BOTTOMLEFT", tab, "BOTTOMLEFT")
			tstrip._borders.bottom:SetPoint("BOTTOMRIGHT", tab, "BOTTOMRIGHT")
			
			tstrip._borders.righttop:SetPoint("BOTTOMRIGHT", tab, "TOPRIGHT")
			tstrip._borders.rightbottom:SetPoint("TOPRIGHT", tab, "BOTTOMRIGHT")
			
			if (tstrip.Event.TabSelected) then
				tstrip.Event.TabSelected(tstrip, tab._value)
			end
		else
			tab:SetBackgroundColor(rgba(clrTransparent))
		end
	end
end

local function _getTabSelectedTS(tab)
	return tab._selected
end


local function _leftDownTS(tab)
	if (not tab._selected) then
		tab:SetSelected(true)
	end
end

local function _addTabTS(tstrip, tabdata)
	if (tabdata) then
		local tab = myUI:CreateFrame("Frame", tstrip)
		
		tab._selected = false
		tab._value = tabdata
		
		tab.SetSelected = _selectTabTS
		tab.GetSelected = _getTabSelectedTS
		
		tab.Event.MouseIn = _mouseInTS
		tab.Event.MouseOut = _mouseOutTS
		tab.Event.LeftDown = _leftDownTS
		
		table.insert(tstrip._tabs, tab)
		
		if (#tstrip._tabs == 1) then
			tab:SetPoint("TOPLEFT", tstrip, "TOPLEFT")
			tab:SetPoint("TOPRIGHT", tstrip, "TOPRIGHT")
			
			tab:SetSelected(true)
		else
			local ptab = tstrip._tabs[#tstrip._tabs - 1]
			tab:SetPoint("TOPLEFT", ptab, "BOTTOMLEFT")
			tab:SetPoint("TOPRIGHT", ptab, "BOTTOMRIGHT")
		end
		
		local txt = myUI:CreateFrame("Text", tab)		
		
		if (tabdata.tostring) then
			txt:SetText(tabdata:tostring())
		else			
			txt:SetText(tostring(tabdata))
		end
		
		tab:SetHeight(txt:GetHeight() + 16)
		txt:SetPoint("LEFTCENTER", tab, "LEFTCENTER", 4, 0)
		
		
		
		return tab
	end
end


local function _bordersTS(tstrip)
	return pairs(tstrip._borders)
end
local function _setBorderSizeTS(tstrip, size)
	if (size) and (size > 0) then
		tstrip._borders.left:SetWidth(size)
		tstrip._borders.top:SetHeight(size)
		tstrip._borders.bottom:SetHeight(size)
		tstrip._borders.righttop:SetWidth(size)
		tstrip._borders.rightbottom:SetWidth(2)
		
		tstrip._borderSize = size
	end
end
local function _getBorderSizeTS(tstrip)
	return tstrip._borderSize
end
local function _setBorderColorTS(tstrip, r, g, b, a)
	for _, border in pairs(tstrip._borders) do
		border:SetBackgroundColor(r,g,b,a)
	end
end

local function _getBorderColorTS(tstrip)
	return tstrip._borders.left:GetBoackgroundColor()
end

local function _createTabstrip(parent)
	local f = myUI:CreateFrame("Frame", parent)
	
	
	f._tabs = {}
	
	f._borders = {
		left = myUI:CreateFrame("Frame", f),
		top = myUI:CreateFrame("Frame", f),
		bottom = myUI:CreateFrame("Frame", f),
		righttop = myUI:CreateFrame("Frame", f),
		rightbottom = myUI:CreateFrame("Frame", f),
	}
	for _, b in pairs(f._borders) do b:SetBackgroundColor(rgba(clrBorder)); b:SetLayer(2) end
	
	f._borders.righttop:SetPoint("TOPRIGHT", f, "TOPRIGHT")
	f._borders.rightbottom:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT")
	
	f.AddTab = _addTabTS
	f.Tabs = _tabsTS
	f.Borders = _bordersTS
	f.SetBorderSize = _setBorderSizeTS
	f.GetBorderSize = _getBorderSizeTS
	f.SetBorderColor = _setBorderColorTS
	f.GetBorderColor = _getBorderColorTS
	
	f.Event = {
		TabSelected = nil -- function(tstrip, tabValue)
	}
	
	f:SetBorderSize(2)
	return f
end

-- Tabpage widget --

local function _bordersTP(tpage)
	return pairs(tpage._borders)
end

local function _setBorderColorTP(tpage, r, g, b, a)
	for _, border in pairs(tpage._borders) do
		border:SetBackgroundColor(r,g,b,a)
	end
end
local function _getBorderColorTP(tpage)
	return tpage._borders.top:GetBackgroundColor()
end
local function _setBorderSizeTP(tpage, size)
	if (size) and (size > 0) then
		tpage._borders.top:SetHeight(size)
		tpage._borders.bottom:SetHeight(size)
		tpage._borders.right:SetWidth(size)
		
		tpage._borderSize = size
	end
end
local function _getBorderSizeTP(tpage)
	return tpage._borderSize
end
local function _createTabpage(parent)
	local f = myUI:CreateFrame("Frame", parent)
	
	f:SetBackgroundColor(rgba(clrActiveTab))
	
	f._borders = {
		top = myUI:CreateFrame("Frame", f),
		right = myUI:CreateFrame("Frame", f),
		bottom = myUI:CreateFrame("Frame", f)
	}
	
	for _, b in pairs(f._borders) do b:SetBackgroundColor(rgba(clrBorder)); b:SetLayer(2) end
	
	f._borders.top:SetPoint("TOPLEFT", f, "TOPLEFT")
	f._borders.top:SetPoint("TOPRIGHT", f, "TOPRIGHT")
	
	f._borders.bottom:SetPoint("BOTTOMLEFT", f, "BOTTOMLEFT")
	f._borders.bottom:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT")
	
	f._borders.right:SetPoint("TOPRIGHT", f, "TOPRIGHT")
	f._borders.right:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT")
	
	f.Borders = _bordersTP
	f.SetBorderColor = _setBorderColorTP
	f.GetBorderColor = _getBorderColorTP
	f.SetBorderSize = _setBorderSizeTP
	f.GetBorderSize = _getBorderSizeTP
	
	f:SetBorderSize(2)
	
	return f
end


-- Color picker widget --

local function _onSlider_CPicker(slider)
	local picker = slider:GetParent()
	
	if (not picker._setting) then
		picker:SetColor(
			picker._rgba.r.slider:GetPosition()/255,
			picker._rgba.g.slider:GetPosition()/255,
			picker._rgba.b.slider:GetPosition()/255,
			picker._rgba.a.slider:GetPosition()/255,
			true
		)
	end
end


local function _getColorCPicker(picker)
	return picker._preview:GetBackgroundColor()
end

local function _setColorCPicker(picker, r, g, b, a, throwEvent)
	picker._setting = true
	
	picker._preview:SetBackgroundColor(r,g,b,a)
	
	local hr = math.floor(r * 255)
	local hg = math.floor(g * 255)
	local hb = math.floor(b * 255)
	local ha = math.floor(a * 255)
	
	picker._rgba.r.value:SetText(tohex(hr))
	picker._rgba.g.value:SetText(tohex(hg))
	picker._rgba.b.value:SetText(tohex(hb))
	picker._rgba.a.value:SetText(tohex(ha))
	
	picker._rgba.r.slider:SetPosition(hr)
	picker._rgba.g.slider:SetPosition(hg)
	picker._rgba.b.slider:SetPosition(hb)
	picker._rgba.a.slider:SetPosition(ha)
	
	if (picker.Event.ColorChanged) and (throwEvent) then picker.Event.ColorChanged(picker, r,g,b,a) end
	
	picker._setting = false
end

local function _createColorPicker(parent)
	local f = myUI:CreateFrame("myUI_BorderBox", parent)
	
	f:SetLayer(parent:GetLayer() + 2)
	
	f._preview = myUI:CreateFrame("myUI_BorderBox", f)
	f._preview:SetSize(48)
	f._preview:SetPoint("TOPLEFT", f, "TOPLEFT", 8, 8)
	
	f._rgba = {r = {}, g = {}, b = {}, a = {}}
	
	for _, t in pairs(f._rgba) do
		t.label = myUI:CreateFrame("Text", f)						
		t.value = myUI:CreateFrame("Text", f)
		
		t.slider = myUI:CreateFrame("RiftSlider", f)		
		t.slider:SetRange(0,255)
		t.slider.Event.SliderChange = _onSlider_CPicker
	end
	
	f._rgba.r.label:SetText(L["red"])
	f._rgba.g.label:SetText(L["green"])
	f._rgba.b.label:SetText(L["blue"])
	f._rgba.a.label:SetText(L["alpha"])
	
	f._rgba.r.label:SetPoint("TOPLEFT", f, "TOPLEFT", 40 + f._preview:GetWidth(), 8)
	f._rgba.g.label:SetPoint("TOPRIGHT", f._rgba.r.label, "BOTTOMRIGHT", 0, 0)
	f._rgba.b.label:SetPoint("TOPRIGHT", f._rgba.g.label, "BOTTOMRIGHT", 0, 0)
	f._rgba.a.label:SetPoint("TOPRIGHT", f._rgba.b.label, "BOTTOMRIGHT", 0, 0)
	
	f._rgba.r.value:SetPoint("TOPLEFT", f, "TOPRIGHT", -32, 8)
	f._rgba.g.value:SetPoint("TOPRIGHT", f._rgba.r.value, "BOTTOMRIGHT", 0, 0)
	f._rgba.b.value:SetPoint("TOPRIGHT", f._rgba.g.value, "BOTTOMRIGHT", 0, 0)
	f._rgba.a.value:SetPoint("TOPRIGHT", f._rgba.b.value, "BOTTOMRIGHT", 0, 0)
	
	
	
	for _, t in pairs(f._rgba) do
			t.value:SetWidth(32)
			t.slider:SetPoint("LEFTCENTER", t.label, "RIGHTCENTER", 12, 6)
			t.slider:SetPoint("RIGHTCENTER", t.value, "LEFTCENTER", -12, 6)
	end
	
	f:SetSize(320, 100)
	
	f.GetColor = _getColorCPicker
	f.SetColor = _setColorCPicker
	
	local function fnil(self) end	-- to prevent click pass through
	f.Event.LeftDown = fnil
	f.Event.RightDown = fnil
	f.Event.MiddleDown = fnil
	f.Event.Mouse4Down = fnil
	f.Event.Mouse5Down = fnil
	
	
	f.Event = {
		ColorChanged = nil -- function(picker, r, g, b, a)
	}
	
	f:SetColor(1,1,1,1)
	
	return f
end





-- Anchorbox widget --

local function _getSelAnchorAB(abox) 
	local a = nil
	for k, v in pairs(lbox._anchors) do
		if (v:GetChecked()) then
			a = v._anchor
			break
		end
	end
	return a
end

local function _setSelAnchorAB(abox, anchor, fromUI)
	local goodAnchor = nil
	local c = false
	
	for k, v in pairs(abox._anchors) do
		c = (anchor == k)
		
		if (not c) then 
			v:SetChecked(false)		
		else
			goodAnchor = k
			if (not _fromUI) then v:SetChecked(true) end	-- would already be checked, so don't reset. That would cause an infinite loop.
		end		
	end	
	
	if (fromUI) and (abox.Event.AnchorChanged) then
		abox.Event.AnchorChanged(abox, goodAnchor)
	end
	
end
local function _clearAB(abox)
	for k, v in pairs(abox._anchors) do
		v:SetChecked(false)
	end
end
local function _anchorClickAB(chk)	
	if (chk:GetChecked()) then
		_setSelAnchorAB(chk:GetParent(), chk._anchor, true)
	end
end

local function _createAnchorBox(parent)
	local f = myUI:CreateFrame("myUI_BorderBox", parent)
	
	local anchors = {"TOPLEFT", "TOPCENTER", "TOPRIGHT", 
		"RIGHTCENTER", "BOTTOMRIGHT", "BOTTOMCENTER", "BOTTOMLEFT", 
		"LEFTCENTER", "CENTER"}
	
	f._anchors = {}
	
	for _, a in ipairs(anchors) do
		local af = myUI:CreateFrame("RiftCheckbox", f)
		af:SetPoint(a, f, a, 0, 0)
		af._anchor = a
		af.Event.CheckboxChange = _anchorClickAB
		
		f._anchors[a] = af
	end
	
	f.GetAnchor = _getSelAnchorAB
	f.SetAnchor = _setSelAnchorAB
	f.ClearAnchors = _clearAB
	
	f.Event = {
		AnchorChanged = nil, 	-- function(abox, anchor)
	}
	
	return f
end

-- Viewport widget --
local function _layoutVP(self)
	local vp = self.Viewport
	local content = self.Content
	local vscroll = self.VScroll
	
	local cycontent = content:GetHeight()
	local cyvp = vp:GetHeight()
	
	if (cycontent > cyvp) then
		local cydelta = cycontent - cyvp
		vscroll:SetRange(0, cydelta)
		vscroll:SetPosition(0)
		vscroll:SetEnabled(true)
	else		
		vscroll:SetEnabled(false)
	end
	content:SetPoint("TOPLEFT", vp, "TOPLEFT")
	content:SetPoint("TOPRIGHT", vp, "TOPRIGHT")
end
local function _contentSizeVP(content)
	_layoutVP(content:GetParent():GetParent())
end

local function _viewportSizeVP(viewport)
	_layoutVP(viewport:GetParent())
end

local function _vscrollVP(vscroll)
	local content = vscroll:GetParent().Content
	local vp = vscroll:GetParent().Viewport
	
	local offy = -vscroll:GetPosition()
	content:SetPoint("TOPLEFT", vp, "TOPLEFT", 0, offy)
	content:SetPoint("TOPRIGHT", vp, "TOPRIGHT", 0, offy)
end

local function _setContentHeightVP(self, cy)
	self.Content:SetHeight(cy)
end

local function _createViewport(parent)
	local f = myUI:CreateFrame("Frame", parent)
	local bbox = myUI:CreateFrame("myUI_BorderBox", f)
	local vp = myUI:CreateFrame("Mask", f)
	local content = myUI:CreateFrame("Frame", vp)
	local vscroll = myUI:CreateFrame("RiftScrollbar", f)
	
	vscroll:SetOrientation("vertical")
	local cxScroll = vscroll:GetWidth()
	
	bbox:SetPoint("TOPLEFT", f, "TOPLEFT")
	bbox:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -cxScroll - 2, 0)
	
	vp:SetPoint("TOPLEFT", bbox, "TOPLEFT", 4, 4)
	vp:SetPoint("BOTTOMRIGHT", bbox, "BOTTOMRIGHT", -4, -4)
	
	vscroll:SetPoint("TOPRIGHT", f, "TOPRIGHT", 0, 4)
	vscroll:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", 0, -4)
	vscroll:SetEnabled(false)
	
	content:SetPoint("TOPLEFT", vp, "TOPLEFT", 0, 0)
	content:SetPoint("TOPRIGHT", vp, "TOPRIGHT", 0, 0)
	content:SetHeight(100) 
	
	f.Viewport = vp
	f.Content = content
	f.VScroll = vscroll
	f.BorderBox = bbox
	
	content.Event.Size = _contentSizeVP
	vp.Event.Size = _viewportSizeVP
	vscroll.Event.ScrollbarChange = _vscrollVP
	
	f.SetContentHeight = _setContentHeightVP
	
	_layoutVP(f)
	
	return f
end

-- button widget

local function _setVisibleBTN(self, visible)
	self.bg:SetVisible(visible)
end
local function _setEnabledBTN(self, enabled)
	self._enabled = enabled
	
	if (enabled) then		
		self:SetFontColor(rgba(clrNormal))
	else
		self:SetFontColor(rgba(clrDisabled))
	end
end
local function _getEnabledBTN(self)
	return self._enabled
end

local function _mouseInBTN(self)
	if (self._enabled) then
		self:SetBackgroundColor(rgba(clrHover))
	end
end
local function _mouseOutBTN(self)
	self:SetBackgroundColor(rgba(clrButton))
end

local function _leftDownBTN(self)
	if (self._enabled) then
		self:SetBackgroundColor(rgba(clrButtonDown))	
	end
end

local function _leftUpBTN(self)
	if (self._enabled) then
		self:SetBackgroundColor(rgba(clrHover))
		if (self.Event.LeftPress) then self.Event.LeftPress(self) end
	end
end

local function _leftUpoutsideBTN(self)
	if (self._enabled) then
		self:SetBackgroundColor(rgba(clrButton))
	end
end

local function _setTextBTN(self, text)
	if (text) then
		self:_setText(" " .. text .. " ")
	end
end


local function _getParentBTN(self)
	return self.bg:GetParent()
end
local function _createButton(parent)
	local bg = myUI:CreateFrame("Frame", parent)
	local txt = myUI:CreateFrame("Text", bg)
	
	txt:SetFontSize(12); txt.customFontSize = true
	
	txt.bg = bg
	bg:SetPoint("TOPLEFT", txt, "TOPLEFT", -2, -2)
	bg:SetPoint("BOTTOMRIGHT", txt, "BOTTOMRIGHT", 2, 2)
	
	bg:SetBackgroundColor(rgba(clrBorder))
	txt:SetBackgroundColor(rgba(clrActiveTab))
	
	
	
	txt.Event.MouseIn = _mouseInBTN
	txt.Event.MouseOut = _mouseOutBTN
	txt.Event.LeftDown = _leftDownBTN
	txt.Event.LeftUp = _leftUpBTN
	txt.Event.LeftUpoutside = _leftUpoutsideBTN
	
	txt._getParent = txt.GetParent
	txt.GetParent = _getParentBTN
	
	txt._setText = txt.SetText
	txt.SetText = _setTextBTN
	txt.SetEnabled = _setEnabledBTN
	txt.GetEnabled = _getEnabledBTN
	txt.SetVisible = _setVisibleBTN
	
	txt.Event = {
		LeftPress = nil
	}
	
	txt:SetBackgroundColor(rgba(clrButton))
	
	txt:SetEnabled(true)
	return txt
end
-- create widget API --

function myUI:CreateWidget(wtype, parent)
	if (wtype == "myUI_BorderBox") then
		return _createBorderBox(parent)
	elseif (wtype == "myUI_Button") then
		return _createButton(parent)
	elseif (wtype == "myUI_Listbox") then
		return _createListbox(parent)
	elseif (wtype == "myUI_Textbox") then
		return _createTextbox(parent)
	elseif (wtype == "myUI_Checkbox") then
		return _createCheckbox(parent)
	elseif (wtype == "myUI_Tabstrip") then
		return _createTabstrip(parent)
	elseif (wtype == "myUI_Tabpage") then
		return _createTabpage(parent)
	elseif (wtype == "myUI_ColorPicker") then
		return _createColorPicker(parent)
	elseif (wtype == "myUI_DropdownList") then
		return _createDropDownList(parent)
	elseif (wtype == "myUI_Anchorbox") then
		return _createAnchorBox(parent)
	elseif (wtype == "myUI_Viewport") then
		return _createViewport(parent)
	end


end













